/*      file     : RouteConverter/TRouteConverter.C
        project  : TRANSIMS
        authors  : Marcus Rickert
        RCS      : $Id: TRouteConverter.C,v 1.7 2005/03/25 12:28:20 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <CommonTypes.h>
#ifdef LINUX
#include <getopt.h>
#endif
#include <TString.h>
#include <TParser.h>
#include <TRandom.h>

#include <TConfiguration.h>
#include <TEnvironment.h>
#include <TBaseDataManager.h>
#include <TObjectManager.h>
#include <TConfiguration.h>
#include <TSynchronize.h>

#include <CAConfig.h>
#include <RouteConverterConfig.h>
#include <TAccessory.h>
#include <TAccessoryHandler.h>
#include <TRouteplan.h>
#include <TVehicle.h>
#include <TNode.h>
#include <TEdge.h>
#include <TAdjacenceInfo.h>

#include <TrafficDataConfig.h>
#include <TTrafficDataManager.h>
#include <TRouter.h>

#include "TLinkHitStat.h"
#include "TRouteConverter.h"

// ***DEFINES************************************************************************* 

#define ROUTECONVERTER_DEFAULT_DEPARTUREBINSIZE 900L

#define AGE_STATFILE_EXTENSION        ".age"
#define LINKHIT_STATFILE_EXTENSION    ".lhs"

#define MAX_AGE 2048

#define ROUTE_BUFFER_SIZE 16384
#define ROUTECONVERTER_DEFAULT_MAXROUTES      (-1L)
#define ROUTECONVERTER_DEFAULT_MINTIME         0.0
#define ROUTECONVERTER_DEFAULT_MAXTIME         24.0
#define ROUTECONVERTER_DEFAULT_SAMPLE          -1.0
#define ROUTECONVERTER_DEFAULT_PIPE            0L
#define ROUTECONVERTER_DEFAULT_BINARYROUTEPLAN 0L
#define ROUTECONVERTER_DEFAULT_LINKHITBINSIZE  0L

#define DEFAULT_PIPE_NAME                "planner_converter_sync"

// ***METHODS************************************************************************* 

// ===================================================================================
#define ParentClass TBaseObject
// =================================================================================== 

TRouteConverter::TRouteConverter(void)

{
    AgeStatOutput = 0;
    LinkHitStatOutput = 0;
    ReroutingTimeCounts = 0;
    DepartureBins = 0;
    MaxRoutes = INVALID;
    MinTime = 0; 
    MaxTime = 24*3600;
    Random = 0;
    RoutesRead = RoutesSampled = RoutesClipped = 0;
    Sync = 0;
    Router = new TRouter;
}

// =================================================================================== 

TRouteConverter::~TRouteConverter(void)

{
    if (AgeStatOutput)
    {
        WriteAgeStat();
        Print("closing age stat file");
        DeleteObject(AgeStatOutput);
        DeleteObjectArray(ReroutingTimeCounts);
    }    
    if (LinkHitStatOutput)
    {
        WriteLinkHitStat();
        Print("closing link hit stat file");
        DeleteObject(LinkHitStatOutput);
    }    
    DeleteObject(Sync);
    DeleteObject(Random);
    DeleteObject(Router);
    DeleteObject(DataManager);
    DeleteObject(BaseDataManager);
    DeleteObject(ObjectManager);
    Print("Goodbye!");
    DeleteObject(Environment);
    DeleteObjectArray(DepartureBins);
}

// =================================================================================== 

bool
TRouteConverter::Activate(void)

{    
    int CBin;

    Configuration->SetDefault(TConfigEntry(ROUTECONVERTER_GROUP, 
                                           ROUTECONVERTER_NAME_MAXROUTES,
                                           ROUTECONVERTER_DEFAULT_MAXROUTES));
    Configuration->SetDefault(TConfigEntry(ROUTECONVERTER_GROUP, 
                                           ROUTECONVERTER_NAME_MINTIME,
                                           ROUTECONVERTER_DEFAULT_MINTIME)); 
    Configuration->SetDefault(TConfigEntry(ROUTECONVERTER_GROUP, 
                                           ROUTECONVERTER_NAME_MAXTIME,
                                           ROUTECONVERTER_DEFAULT_MAXTIME));
    Configuration->SetDefault(TConfigEntry(ROUTECONVERTER_GROUP, 
                                           ROUTECONVERTER_NAME_SAMPLE,
                                           ROUTECONVERTER_DEFAULT_SAMPLE));
    Configuration->SetDefault(TConfigEntry(ROUTECONVERTER_GROUP, 
                                           ROUTECONVERTER_NAME_DEPARTUREBINSIZE,
                                           ROUTECONVERTER_DEFAULT_DEPARTUREBINSIZE));
    Configuration->SetDefault(TConfigEntry(ROUTECONVERTER_GROUP, 
                                           ROUTECONVERTER_NAME_PIPE,
                                           ROUTECONVERTER_DEFAULT_PIPE));
    Configuration->SetDefault(TConfigEntry(ROUTECONVERTER_GROUP, 
                                           ROUTECONVERTER_NAME_BINARYROUTEPLAN,
                                           ROUTECONVERTER_DEFAULT_BINARYROUTEPLAN));
    Configuration->SetDefault(TConfigEntry(ROUTECONVERTER_GROUP, 
                                           ROUTECONVERTER_NAME_LINKHITBINSIZE,
                                           ROUTECONVERTER_DEFAULT_LINKHITBINSIZE));

    Configuration->Set(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_LOADACCESSORIES, TRAFFIC_LOADACCESSORIES));

    ObjectManager->RegisterClass(CI_FLAG_TABLE, TLinkHitStat::ClassInfo, CLASS_EDGE, 
                                 CLASS_LINKHITSTAT, this);
    if (!(Environment->Activate() AND
          ObjectManager->Activate() AND
          BaseDataManager->Activate() AND
          TrafficDataManager->Activate() AND
          Router->Activate()))
    {
        Error("cannot activate helper classes");
        return FALSE;
    }
    if (!(Environment->Run() AND
          ObjectManager->Run() AND
          BaseDataManager->Run() AND
          TrafficDataManager->Run()))
    {
        Error("cannot run helper classes");
        return FALSE;
    }
    if (!Configuration->Exists(ROUTECONVERTER_GROUP, ROUTECONVERTER_NAME_FILENAME))
    {
        Error("Activate: no routeplan filename given");
        return FALSE;
    }
    RouteplanFilename = Environment->ReplacePathTokens(Configuration->GetString());
    MaxRoutes = Configuration->GetLong(ROUTECONVERTER_GROUP,
                                       ROUTECONVERTER_NAME_MAXROUTES);
    DepartureBinSize = Configuration->GetLong(ROUTECONVERTER_GROUP,
                                              ROUTECONVERTER_NAME_DEPARTUREBINSIZE);
    NrOfDepartureBins = 24 * 3600 / DepartureBinSize + 1;
    DepartureBins = new long [NrOfDepartureBins];
    for (CBin = 0; CBin < NrOfDepartureBins; CBin++)
        DepartureBins[CBin] = 0;
    Sample = Configuration->GetDouble(ROUTECONVERTER_GROUP,
                                      ROUTECONVERTER_NAME_SAMPLE);
    MinTime = (long) (3600 * Configuration->GetDouble(ROUTECONVERTER_GROUP,
                                                      ROUTECONVERTER_NAME_MINTIME));
    MaxTime = (long) (3600 * Configuration->GetDouble(ROUTECONVERTER_GROUP,
                                                      ROUTECONVERTER_NAME_MAXTIME));
    if (Sample != INVALID)
        Random = new TRandom;
    if (Configuration->GetLong(ROUTECONVERTER_GROUP, ROUTECONVERTER_NAME_PIPE)) 
    {
        Sync = new TSynchronize;
        Print("waiting for sync with planner...");
        if (!Sync->Activate(DEFAULT_PIPE_NAME, FALSE))
        {
            Error("cannot synchronize with plannes");
            return FALSE;
        }
        Print("OK");
    }
    Binary = Configuration->GetLong(ROUTECONVERTER_GROUP, 
                                    ROUTECONVERTER_NAME_BINARYROUTEPLAN);
    LinkHitBinSize = Configuration->GetLong(ROUTECONVERTER_GROUP, 
                                            ROUTECONVERTER_NAME_LINKHITBINSIZE);
    return TRUE;
}

/* =================================================================================== */

bool
TRouteConverter::ActivateAgeStat(void)

{
    if (AgeStatOutput == 0)
    {
        char Filename[512];
        int  CTime;

        AgeStatOutput = new TStream;
        sprintf(Filename, "%s%s", RouteplanFilename(), AGE_STATFILE_EXTENSION);
        if (!AgeStatOutput->Open(Filename, "w"))
        {
            Error("cannot open age stat file %s", Filename);
            return FALSE;
        }
        Print("age stat file %s opened", Filename);
        ReroutingTimeCounts = new long[MAX_AGE];
        for (CTime = 0; CTime < MAX_AGE; CTime++)
            ReroutingTimeCounts[CTime] = 0;
    }
    return TRUE;
}

/* =================================================================================== */

bool
TRouteConverter::ActivateLinkHitStat(void)

{
    if (LinkHitStatOutput == 0)
    {
        char Filename[512];

        LinkHitStatOutput = new TStream;
        sprintf(Filename, "%s%s", RouteplanFilename(), LINKHIT_STATFILE_EXTENSION);
        if (!LinkHitStatOutput->Open(Filename, "w"))
        {
            Error("cannot open link hit stat file %s", Filename);
            return FALSE;
        }
        Print("link hit stat file %s opened", Filename);
    }
    return TRUE;
}

// =================================================================================== 

bool
TRouteConverter::Clip(void)

{
    TParser Parser;
    long    RoutesConverted;

    if (Sync)
    {
        bool ErrorFlag;
        long Step = 1;

        Print("waiting for first route from planner...");
        while (Sync->Wait(Step, ErrorFlag))
            sleep(1);
        if (ErrorFlag)
        {
            Error("while waiting for first route");
            return FALSE;
        }
    }
    if (!Router->OpenRouteplanForWriting())
    {
        Error("Convert: cannot open routeplan file for writing");
        return FALSE;
    }
    if (Configuration->Exists(ROUTECONVERTER_GROUP, ROUTECONVERTER_NAME_AGESTAT))
        if (!ActivateAgeStat())
        {
            Error("Clip: cannot activate age stat");
            return FALSE;
        }
    if (LinkHitBinSize > 0)
        if (!ActivateLinkHitStat())
        {
            Error("Clip: cannot activate link hit stat");
            return FALSE;
        }
    if (!Parser.Open(RouteplanFilename, "r", Binary))
    {
        Error("cannot open TRANSIMS input file '%s'", RouteplanFilename());
        return FALSE;
    }
    Print("%splanner input file '%s' opened", Binary? "binary ":"", RouteplanFilename());
    if (!Binary)
        if (!Parser.ReadToken())
        {
            Warning("empty TRANSIMS input file");
            return TRUE;
        }
    Print("converting route plans (time window %d:%02d-%d:%02d)",
          MinTime / 3600, (MinTime / 60) % 60,
          MaxTime / 3600, (MaxTime / 60) % 60);
    RoutesRead = 0;
    RoutesClipped = 0;
    RoutesSampled = 0;
    MinTimeFound = 24*3600;
    MaxTimeFound = 0;
    RoutesConverted = 0;
    while (Sync OR
           ((MaxRoutes == INVALID OR RoutesClipped < MaxRoutes) AND
            !Parser.EndOfFile()))
    {
        if (Sync AND RoutesConverted <= RoutesRead)
        {
            bool Error;
            
            RoutesConverted = RoutesRead + 1;
            while (Sync->Wait(RoutesConverted, Error))
                sleep(1);
            if (Error)
            {
                Warning("while waiting for route %ld", RoutesConverted);
                break;
            }
            if (RoutesConverted == 12345678)
            {
                Print("last route received");
                break;
            }
            Parser.ClearError();
        }
        if (Sync == 0 AND RoutesRead % 1000 == 0)
            Print("%ld%% of file read", Parser.GetPercentage());
        if (!ClipRoute(&Parser))
        {
            if (Parser.EndOfFile())
                break;
            return FALSE;
        }
    }
    if (Sync)
        Sync->Step(1);
    Parser.Close();
    Print("planner input file closed");
    Print("%ld out of %ld routes clipped", RoutesClipped, RoutesRead);
    if (RoutesClipped > 0)
        Print("routes are in time interval %d:%02d-%d:%02d",
              MinTimeFound / 3600, MinTimeFound / 60 % 60,
              MaxTimeFound / 3600, MaxTimeFound / 60 % 60);
    return TRUE;
}

// =================================================================================== 

bool
TRouteConverter::ClipRoute(TParser *Parser)

{
    long              PlanNumber, PlanMode;
    LongID            VehicleID;
    long              DriverPlanNumber, NrOfTransimsLinks;
    LongID            SourceAccessory, DestAccessory;
    LongID            SourceID, DestID;
    long              CEdge;
    long              DepartureTime;
    TRouteplan       *Routeplan = 0;
    TVehicle         *Vehicle = 0;
    int               NrOfPaminaLinks;
    int               FirstPaminaLinkOffset;
    short int         RouteplanFlags;
    TEdge            *PaminaLinks[MAX_ROUTE_LENGTH];
    TRouteStep        Entries[MAX_ROUTE_LENGTH];
    int               Bin;
    
    if (!(Parser->Decode(PlanNumber) AND
          Parser->Decode(PlanMode) AND
          Parser->Decode(VehicleID) AND
          Parser->Decode(DriverPlanNumber) AND
          Parser->Decode(SourceAccessory) AND
          Parser->Decode(NrOfTransimsLinks)))
    {
        Error("cannot read TRANSIMS route header");
        return FALSE;
    }
    if (ReroutingTimeCounts AND PlanMode < MAX_AGE)
        ReroutingTimeCounts[PlanMode]++;
    if (NrOfTransimsLinks > MAX_ROUTE_LENGTH)
    {
        Error("length (%ld) of route %ld excedes %ld", NrOfTransimsLinks, 
              PlanNumber, MAX_ROUTE_LENGTH);
        return FALSE;
    }
    for (int i = 0; i < NrOfTransimsLinks; i++)
        if (!(Parser->Decode(Entries[i].ID) &&
              Parser->Decode(Entries[i].Estimate)))
        {
            Error("cannot read links of plan # %ld", PlanNumber);
            return FALSE;
        }
    if (!(Parser->Decode(DestAccessory) AND
          Parser->Decode(Entries[NrOfTransimsLinks].Estimate)))
    {
        Error("cannot read destination of plan # %ld", PlanNumber);
        return FALSE;
    }
    RoutesRead++;
    if (Random AND 100.0 * Random->Double() > Sample)
        return TRUE;    
    RoutesSampled++;
    if (!Router->ClipTransimsRoute2Pamina(PlanNumber, MinTime, MaxTime,
                                          SourceAccessory, DestAccessory,
                                          Entries, NrOfTransimsLinks,
                                          PaminaLinks, NrOfPaminaLinks,
                                          FirstPaminaLinkOffset, RouteplanFlags,
                                          SourceID, DestID))
        return TRUE;
    DepartureTime = Entries[FirstPaminaLinkOffset].Estimate;
    if (DepartureTime < MinTimeFound)
        MinTimeFound = DepartureTime;
    if (DepartureTime > MaxTimeFound)
        MaxTimeFound = DepartureTime;
    Bin = DepartureTime / DepartureBinSize;
    if (Bin < NrOfDepartureBins)
        DepartureBins[Bin]++;
    Routeplan = new TRouteplan(Entries[FirstPaminaLinkOffset].Estimate, RouteplanFlags, 
                               SourceID, DestID, PlanNumber);
    for (CEdge = 0; CEdge < NrOfPaminaLinks; CEdge++)
    {
        if (LinkHitBinSize)
            Hit(PaminaLinks[CEdge]->GetParallelID(), Entries[FirstPaminaLinkOffset + CEdge].Estimate);
        Routeplan->AddStep(PaminaLinks[CEdge]->GetParallelID(), 
                           Entries[FirstPaminaLinkOffset + CEdge + 1].Estimate);
    }
    Vehicle = new TVehicle(PlanNumber);
    Vehicle->SetRouteplan(Routeplan);
    Vehicle->SetMaxVelocity(CA_VELOCITY_TO_REAL_VELOCITY(CA_DEFAULT_V_MAX));
    Vehicle->SetCurrentVelocity(CA_VELOCITY_TO_REAL_VELOCITY(CA_DEFAULT_V_MAX));
    if (!Router->WriteRoute(Vehicle))
        return FALSE;
    DeleteObject(Vehicle);
    RoutesClipped++;
    return TRUE;
}

// =================================================================================== 

TString
TRouteConverter::GetClassName(void) const

{
    return "TRouteConverter";
}

// =================================================================================== 

TString
TRouteConverter::GetInstanceName(void) const

{
    char Line[256];

    sprintf(Line, "%ld/%ld/%ld routes read/sampled/clipped", 
            RoutesRead, RoutesSampled, RoutesClipped);
    return Line;
}

// =================================================================================== 

bool
TRouteConverter::HandleCreateRequest(TParallelObject *&Object, 
                                     const TClassInfo *ClassInfo, 
                                     TParallelID ParallelID)

{
    if (ClassInfo == TLinkHitStat::ClassInfo)
    {   
        Object = new TLinkHitStat(ParallelID, MinTime, MaxTime, LinkHitBinSize);
    }
    return TRUE;
}

// =================================================================================== 

bool
TRouteConverter::HandleLoadRequest(TParallelObject *Object, bool &Loaded)

{
    if (Object->GetClassID() == TLinkHitStat::ClassInfo->GetID())
    {
        Loaded = TRUE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TRouteConverter::Hit(TParallelID EdgeID, long TimeStep)

{
    TEdge        *Edge;
    TLinkHitStat *Stat;
    TInfoID       InfoID;

    if ((Edge = TEdge::GetPtr(EdgeID, TRUE)) == 0)
    {
        Error("Hit: cannot find edge %ld", EdgeID);
        return FALSE;
    }
    InfoID = TLinkHitStat::ClassInfo->GetInfoID();
    if ((Stat = (TLinkHitStat *) Edge->GetInfo(InfoID, TRUE, TRUE)) == 0)
    {
        Error("Hit: cannot retrieve/create hit link stat for %s",
              (char *) Edge->GetObjectName());
        return FALSE;
    }
    Stat->Hit(TimeStep);
    return TRUE;
}

// =================================================================================== 

bool
TRouteConverter::ReadNetwork(void)

{
    TInfoMask EdgeLoadMask;

    if (!DataManager->CreateSimulationObjects(&Nodes, &Edges))
    {
        Error("cannot create simulation objects");
        return FALSE;
    }
    EdgeLoadMask = TAdjacenceInfo::ClassInfo->GetInfoMask();
    if (Configuration->GetLong(TRAFFIC_GROUP, TRAFFIC_NAME_LOADACCESSORIES))
        EdgeLoadMask |= TAccessoryHandler::ClassInfo->GetInfoMask();
    if (!ObjectManager->FillContainers((TContainerObjectList *) &Edges, EdgeLoadMask))
    {
        Error("cannot fill nodes and edges");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TRouteConverter::Start(int argc, char *argv[])

{
    Initialize();
    Environment = new TEnvironment(argc, argv);
    ObjectManager = new TObjectManager;
    BaseDataManager = new TBaseDataManager;
    DataManager = new TTrafficDataManager;
    Print("Hello World!");
    if (! (Activate() AND
           ReadNetwork() AND
           Clip() AND
           WriteStatistics() AND
           Router->CloseFiles()))
    {
        Error("during execution");
        return FALSE;
    }
    Print("success");
    return TRUE;
}

// =================================================================================== 

bool
TRouteConverter::WriteAgeStat(void)

{
    double SumAge;
    long   CTime;
    long   Count;
    long   MaxTime;

    Count = 0;
    SumAge = 0.0;
    MaxTime = 0;
    for (CTime = 0; CTime < MAX_AGE; CTime++)
        if (ReroutingTimeCounts[CTime] > 0)
            MaxTime = CTime;
    for (CTime = 0; CTime <= MaxTime; CTime++)
        if (ReroutingTimeCounts[CTime] > 0)
        {
            long Age;

            Age = MaxTime - CTime;
            Count += ReroutingTimeCounts[CTime];
            SumAge += 1.0 * Age * ReroutingTimeCounts[CTime];
        }
    if (Count > 0)
    {
        AgeStatOutput->Write("# max age:           %ld\n", MaxTime);
        AgeStatOutput->Write("# average age:       %f\n", SumAge / Count);
        AgeStatOutput->Write("# not yet re-routed: %f\n",
                             1.0 * ReroutingTimeCounts[0] / Count);
        for (CTime = 0; CTime <= MaxTime; CTime++)
            if (ReroutingTimeCounts[CTime] > 0)
                AgeStatOutput->Write("%ld %ld %ld %f\n", MaxTime - CTime, CTime,
                                     ReroutingTimeCounts[CTime],
                                     1.0 * ReroutingTimeCounts[CTime] / Count);
    }
    return TRUE;
}

// =================================================================================== 

bool
TRouteConverter::WriteLinkHitStat(void)

{
    TEnumerator< TLinkHitStat > CStat(TLinkHitStat::ClassInfo->GetTable());
    
    CStat.Enumerate();
    while (CStat())
    {
        if (!CStat->Write(LinkHitStatOutput))
        {
            Error("while writing hit link stat");
            return FALSE;
        }
        CStat++;
    }    
    return TRUE;
}

// =================================================================================== 

bool
TRouteConverter::WriteRoute(const TTimeStep &, TVehicle *Vehicle)

{
    RoutesClipped++;
    return Router->WriteRoute(Vehicle);
}

// =================================================================================== 

bool
TRouteConverter::WriteStatistics(void)

{
    int CBin;

    Router->WriteStat("#--conversion statistics (TRouteConverter) --------------------");
    Router->WriteStat("#%ld out of %ld routes clipped", RoutesClipped, RoutesRead);
    for (CBin = 0; CBin < NrOfDepartureBins; CBin++)
        if (DepartureBins[CBin] > 0)
            Router->WriteStat("%ld %f %ld", CBin * DepartureBinSize,
                              1.0 * CBin * DepartureBinSize / 3600.0, DepartureBins[CBin]);
    return Router->WriteStatistics();
}

// =================================================================================== 
#undef ParentClass


