/*      file     : XTrafficData/TMapView.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TMapView.C,v 1.6 2005/04/03 20:46:57 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <Graphx.h>
#include <XGraphx.h>

#include <TConfiguration.h>
#include <TParallel.h>
#include <TLocationInfo.h>
#include <TNode.h>
#include <TEdge.h>
#include <TEdgeInfo.h>
#include <TGraphEdgeInfo.h>
#include <TSamplingDescription.h>
#include <TAggregationInfo.h>
#include <TAggregationResult.h>
#include <TStatManager.h>

#include <Transims.h>
#include <TrafficDataTypes.h>
#include <TMultilaneDump.h>
#include <TGridStatLookup.h>

#include <XTrafficDataConfig.h>

#include "TMapViewer.h"
#include "TMapView.h"

// ***DEFINES************************************************************************* 

#define MAPWIN_DEFAULT_BACKGROUNDCOLOR 220L
#define MAPWIN_DEFAULT_LINECOLOR       0L
#define MAPWIN_DEFAULT_BLOCKCOLOR      0L
#define MAPWIN_DEFAULT_VEHICLECOLOR    250L
#define MAPWIN_DEFAULT_ABSORPTIONCOLOR GR_RED
#define MAPWIN_DEFAULT_EMISSIONCOLOR   GR_GREEN
#define MAPWIN_DEFAULT_DUMPINTERVAL    1L

#define MAPWIN_FONT_NAME  "Courier-Medium-Regular-12"
#define MAPWIN_RESERVED_COLORS 8

#define HIGH_DENS_COLOR "red"
#define LOW_DENS_COLOR  "white"

#define REROUTED_FRAME_COLOR 255

// ***CONSTANTS*********************************************************************** 

int DelayColors[8] =

{
    0,        // state unknown
    GR_GREEN,  // very early
    GR_GREEN,
    GR_CYAN,
    GR_BLUE,       // little delay
    GR_YELLOW,
    GR_ORANGE,
    GR_RED     // very late
};
    
    
// ***METHODS************************************************************************* 

//FILE *output;

// =================================================================================== 
#define ParentClass TWindow
// =================================================================================== 

TMapView::TMapView(TMapViewer *theMapViewer, const char *Name, TRectangle *aFrame, 
                   TEdgeList *theEdges, int theWidth, int theHeight) :

ParentClass(Name, theWidth, theHeight, MAPWIN_RESERVED_COLORS)

{
    MapViewer = theMapViewer;
    Frame = *aFrame;
    Edges = theEdges;
    VisibleEdges = new TEdgeList;
    VehicleRadius = 0.55 * SITE_LENGTH_IN_CM;

    Configuration->SetDefault(TConfigEntry(MAPWIN_GROUP, MAPWIN_NAME_DUMPINTERVAL, MAPWIN_DEFAULT_DUMPINTERVAL));

    BackgroundColor = Configuration->Exists(MAPWIN_GROUP, MAPWIN_NAME_BACKGROUNDCOLOR)? 
        Configuration->GetLong() : MAPWIN_DEFAULT_BACKGROUNDCOLOR;
    LineColor = Configuration->Exists(MAPWIN_GROUP, MAPWIN_NAME_LINECOLOR)? 
        Configuration->GetLong() : MAPWIN_DEFAULT_LINECOLOR;
    VehicleColor = Configuration->Exists(MAPWIN_GROUP, MAPWIN_NAME_VEHICLECOLOR)? 
        Configuration->GetLong() : MAPWIN_DEFAULT_VEHICLECOLOR;
    BlockColor = Configuration->Exists(MAPWIN_GROUP, MAPWIN_NAME_BLOCKCOLOR)? 
        Configuration->GetLong() : MAPWIN_DEFAULT_BLOCKCOLOR;
    AbsorptionColor = Configuration->Exists(MAPWIN_GROUP, MAPWIN_NAME_ABSORPTIONCOLOR)? 
        Configuration->GetLong() : MAPWIN_DEFAULT_ABSORPTIONCOLOR;
    EmissionColor = Configuration->Exists(MAPWIN_GROUP, MAPWIN_NAME_EMISSIONCOLOR)? 
        Configuration->GetLong() : MAPWIN_DEFAULT_EMISSIONCOLOR;
    DumpInterval = Configuration->GetLong(MAPWIN_GROUP, MAPWIN_NAME_DUMPINTERVAL);
    AllocateColorTransition(LOW_DENS_COLOR, HIGH_DENS_COLOR, MAPWIN_RESERVED_COLORS, ColorBase);
    BoxLineColor = 0;
    SetXaxis(Frame.GetMinX(), Frame.GetMaxX());
    SetYaxis(Frame.GetMinY(), Frame.GetMaxY());
    SetVerticalTextAlignment(GrTopAlign);
    X11SetSpeed(GrFast);
    X11WaitUntilExposed();
    FrameChanged = TRUE;
    Mode = LineViewMode;
    MarkActive = FALSE;
    DumpDesc = 0;
    DumpInfo = 0;
    if ((Font = LoadFontName(MAPWIN_FONT_NAME)) == 0)
        Fail("cannot allocate font");


    //    output = fopen("edges", "w");
}

// =================================================================================== 

TMapView::~TMapView(void)

{
    //    fclose(output);
    DeleteObject(VisibleEdges);
    StopMultilaneDump();
}

// =================================================================================== 

bool
TMapView::CheckVisibleEdges(void)

{    
    TEdgeEnum CEdge;
    TInfoID   InfoID;

    VisibleEdges->DeleteAll();
    CEdge.WalksForwardThrough(Edges);
    InfoID = TGraphEdgeInfo::ClassInfo->GetInfoID();
    while (CEdge())
    {
        TGraphEdgeInfo *GraphEdge;

        if ((GraphEdge = (TGraphEdgeInfo *) CEdge->GetInfo(InfoID, TRUE, TRUE)) == 0)
        {
            Error("DrawMap: cannot retrieve graph info for %s",
                  (char *) CEdge->GetObjectName());
            return FALSE;
        }
        if (Frame.Contains(GraphEdge->From) OR
            Frame.Contains(GraphEdge->To))
            VisibleEdges->Append(CEdge);
        CEdge++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TMapView::Draw(void)

{
    if (Mode == DensityViewMode)
        ClearScreen(255);
    else
        ClearScreen(BackgroundColor);
    if (FrameChanged)
    {
        if (!CheckVisibleEdges())
        {
            Error("Draw: while checking visible edges");
            return FALSE;
        }
        FrameChanged = FALSE;
    }
    if (!DrawMap())
        return FALSE;
    DrawMisc();
    return ParentClass::Draw();
}

// =================================================================================== 

void
TMapView::DrawColoredEdge(TGraphEdgeInfo *GraphEdge, int Color)

{
    TLocation Vector;
    TLocation Normal;
    double    Length;

    Vector = GraphEdge->To - GraphEdge->From;
    Length = Vector.Length();
    Vector *= SITE_LENGTH_IN_CM / Length;
    Normal = Vector.Normal();    
    * this << GrLocLine(GraphEdge->From + 1.0 * Normal, 
                        GraphEdge->To + 1.0 * Normal, Color);    
}

// =================================================================================== 

bool
TMapView::DrawDensity(TEdge *Edge, TGraphEdgeInfo *GraphEdge)

{
    TGridStatLookup *Lookup;
    TEdgeInfo       *EdgeInfo;
    double           Density;
    int              Color = GR_VIOLET;

    if (!GetGraphicData(Edge, EdgeInfo, Lookup))
    {
        Warning("DrawDensity: cannot retrieve graphics info for %s:",
                (char *) Edge->GetObjectName());
        return FALSE;
    }
    Density = Lookup->GetAverageDensity(DisplayTime);
    if (Density < 0.1)
        Color = 200; // grey
    else if (Density < 0.3)
        Color = GR_GREEN;
    else if (Density < 0.5)
        Color = GR_YELLOW;
    else if (Density < 0.7)
        Color = GR_RED;
    else
        Color = GR_VIOLET;
    DrawColoredEdge(GraphEdge, Color);
    return TRUE;
}

// =================================================================================== 

bool
TMapView::DrawDump(TEdge *Edge, TGraphEdgeInfo *Info)

{
    TMultilaneDump *DumpData;
    TLocation       Vector;
    TLocation       EffVector;
    TLocation       Normal;
    double          Length;
    int             CLane, CEntry;

    // Draw NrOfLanes+1 lane separators and boundaries

    *this << GrLocLine(Info->From, Info->To, LineColor);

    if ((DumpData = (TMultilaneDump *) 
         Edge->GetInfo(TMultilaneDump::ClassInfo->GetInfoID())) == 0)
        return TRUE;

    // compute a unit vector pointing from 'From' location to 'To' Location with length
    // of SITE_LENGTH_IN_CM 

    Vector = Info->To - Info->From;
    EffVector = Vector;

    Length = Vector.Length();

    Vector *= SITE_LENGTH_IN_CM / Length;
    EffVector *= SITE_LENGTH_IN_CM / Info->EffectiveLength;

    Normal = Vector.Normal();
    
    *this << GrLocLine(Info->From + 0.5 * Normal, 
                       Info->From + (DumpData->GetNrOfLanes() + 0.5) * Normal,
                       LineColor);
    *this << GrLocLine(Info->To + 0.5 * Normal, 
                       Info->To + (DumpData->GetNrOfLanes() + 0.5) * Normal,
                       LineColor);

    if (DumpData->GetNrOfLanes() > 0)
        for (CLane = 0; CLane <= DumpData->GetNrOfLanes(); CLane++)
        {
            TLocation Offset ((0.5 + CLane) * Normal);
            
            (*this) << GrLocLine(Offset + Info->From, Offset + Info->To, LineColor);
        }
    for (CLane = 0; CLane < DumpData->GetNrOfLanes(); CLane++)
        for (CEntry = 0; CEntry < DumpData->GetNrOfEntries(CLane); CEntry++)
        {
            long Position = DumpData->GetEntry(CLane, CEntry).Position;
            TSite Site = DumpData->GetEntry(CLane, CEntry).Site;
            
            TLocation Loc (Info->From + 1.0 * Position * EffVector + 
                           (DumpData->GetNrOfLanes() - CLane) * Normal);
            
            if (Site & SITE_REROUTED_VEHICLE)
                (*this) << GrFillCircle(Loc.x(), Loc.y(), 1.4 * VehicleRadius, 
                                        REROUTED_FRAME_COLOR);
            if (SITE_IS_VEHICLE(Site))
            {
                int VColor;

                if (Mode == DelayViewMode)
                    VColor = DelayColors[SITE_GET_DELAY(Site)];
                else
                    VColor = 255;
                (*this) << GrFillCircle(Loc.x(), Loc.y(), VehicleRadius, VColor);
                //              Print("draw vehicle with color %d", VColor);
            }
            if (Site & SITE_ABSORPTION)
                (*this) << GrCircle(Loc.x(), Loc.y(), VehicleRadius, AbsorptionColor);
            if (Site & SITE_EMISSION)
                (*this) << GrCircle(Loc.x(), Loc.y(), VehicleRadius, EmissionColor);
            if (Site & SITE_BLOCKED)
                (*this) << GrFillCircle(Loc.x(), Loc.y(), VehicleRadius, 0);
        }
    char IdString[20];

    sprintf(IdString, "%d", Edge->GetFromNode()->GetParallelID());
    WriteCentered(Info->From.x(), Info->From.y(), IdString, GR_RED);

    sprintf(IdString, "%d", Edge->GetToNode()->GetParallelID());
    WriteCentered(Info->To.x(), Info->To.y(), IdString, GR_RED);
    return TRUE;
}

// =================================================================================== 

bool
TMapView::DrawLine(TEdge *, TGraphEdgeInfo *GraphEdge)

{
    if (GraphEdge->FromColor == GraphEdge->ToColor)
        * this << GrLocLine(GraphEdge->From, GraphEdge->To, GraphEdge->FromColor);
    else
    {
        TLocation Middle (0.5 * (GraphEdge->From + GraphEdge->To));
        
        * this << GrLocLine(GraphEdge->From, Middle, GraphEdge->FromColor);
        * this << GrLocLine(Middle, GraphEdge->To, GraphEdge->ToColor);
    }

    //    fprintf(output, "%f %f\n%f %f\n\n", GraphEdge->From.x(), GraphEdge->From.y(),
    //      GraphEdge->To.x(), GraphEdge->To.y());

    return TRUE;
}

// =================================================================================== 

bool
TMapView::DrawMap(void)

{
    TEdgeEnum CEdge;
    TInfoID   InfoID;

    CEdge.WalksForwardThrough(VisibleEdges);
    InfoID = TGraphEdgeInfo::ClassInfo->GetInfoID();
    while (CEdge())
    {
        TGraphEdgeInfo *GraphEdge;

        if ((GraphEdge = (TGraphEdgeInfo *) CEdge->GetInfo(InfoID, TRUE, TRUE)) == 0)
        {
            Error("DrawMap: cannot retrieve graph info for %s",
                  (char *) CEdge->GetObjectName());
            return FALSE;
        }
        switch (Mode)
        {
        case LineViewMode:

            DrawLine(CEdge, GraphEdge);
            break;

        case VelocityViewMode:

            DrawVelocity(CEdge, GraphEdge);
            break;

        case DensityViewMode:

            DrawDensity(CEdge, GraphEdge);
            break;

        case DumpViewMode:
        case DelayViewMode:

            DrawDump(CEdge, GraphEdge);
            break;

        default:

            Error("DrawMap: ViewMode not supported");
            return FALSE;
        }
        CEdge++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TMapView::DrawMisc(void)

{
    TMapViewEnum  CView;

    if (MarkActive)
        *this << GrLocPoint(FirstMark, GR_RED);
    CView.WalksForwardThrough(MapViewer->GetMapViewList());
    while (CView())
    {
        *this << GrRectangle(CView->GetFrame(), BoxLineColor);
        CView++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TMapView::DrawVelocity(TEdge *Edge, TGraphEdgeInfo *GraphEdge)

{
    TGridStatLookup *Lookup;
    TEdgeInfo       *EdgeInfo;
    double           Velocity, Ratio;
    int              Color = GR_VIOLET;

    if (!GetGraphicData(Edge, EdgeInfo, Lookup))
    {
        Warning("DrawVelocity: cannot retrieve graphicsinfo for %s:",
                (char *) Edge->GetObjectName());
        return FALSE;
    }
    Velocity = Lookup->GetAverageVelocity(DisplayTime);
    if (Velocity == 0.0)
        Color = GR_VIOLET;
    else
        if (Velocity == (double) INVALID)
            Color = ColorBase;
        else
        {
            Ratio = KMH2MPS(CA_VELOCITY_TO_REAL_VELOCITY(Velocity)) / EdgeInfo->SpeedLimit;
            Color = ColorBase + TransimsLinkColor(Ratio, MAPWIN_RESERVED_COLORS);
        }
    DrawColoredEdge(GraphEdge, Color);
    return TRUE;
}

// =================================================================================== 

TString
TMapView::GetClassName(void) const

{
    return "TMapView";
}

// =================================================================================== 

bool
TMapView::GetGraphicData(TEdge *Edge, TEdgeInfo *&EdgeInfo, 
                         TGridStatLookup *&GridStatLookup)

{
    TInfoID GridStatInfoID, EdgeInfoID;

    EdgeInfoID = TEdgeInfo::ClassInfo->GetInfoID();
    if ((EdgeInfo = (TEdgeInfo *) Edge->GetInfo(EdgeInfoID, TRUE, TRUE)) == 0)
    {
        Warning("GetGraphicData: cannot retrieve edge info for %s:",
                (char *) Edge->GetObjectName());
        return FALSE;
    }
    GridStatInfoID = GridStatLookupClassInfo->GetInfoID();
    if ((GridStatLookup = (TGridStatLookup *) Edge->GetInfo(GridStatInfoID)) == 0)
    {
        Warning("GetGraphicData: cannot retrieve grid stat for %s",
                (char *) Edge->GetObjectName());
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

const TRectangle &
TMapView::GetFrame(void) const

{
    return Frame;
}

// =================================================================================== 

TString
TMapView::GetInstanceName(void) const

{
    char Line[256];

    sprintf(Line, "visible:%ld", VisibleEdges->GetNrOfElements());
    return Line;
}

// =================================================================================== 

bool
TMapView::HandleAggregationResult(TAggregationInfo *AggregationInfo, 
                                  TTimeStep &TimeStep)

{
    if (AggregationInfo == DumpInfo)
        return HandleDumpResult(AggregationInfo, TimeStep);
    return TStatHandler::HandleAggregationResult(AggregationInfo, TimeStep);
}

// =================================================================================== 

bool
TMapView::HandleDumpResult(TAggregationInfo *AggregationInfo, TTimeStep &TimeStep)

{
    TAggregationResult *Result = 0;
    TMultilaneDumpList *DumpList = 0;
    TMultilaneDumpEnum  CDump;
    TInfoID             InfoID;

    if ((Result = (TAggregationResult *) AggregationInfo->GetResult(TimeStep)) == 0)
    {
        Error("HandleDumpResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((DumpList = (TMultilaneDumpList *) Result->StatObjectList) == 0)
    {
        Warning("HandleDumpResult: cannot retrieve stat list");
        return TRUE;
    }
    InfoID = TMultilaneDump::ClassInfo->GetInfoID();
    CDump.WalksForwardThrough(DumpList);
    while (CDump())
    {   
        TEdge *Edge;

        if ((Edge = (TEdge *) TEdge::ClassInfo->GetObjectPtr(CDump->GetParallelID())) == 0)
        {
            Error("HandleDumpResult: cannot find edge for %s",
                  (char *) CDump->GetObjectName());
            return FALSE;
        }
        Edge->SetInfo(InfoID, (TMultilaneDump *) CDump, TRUE);
        CDump++;
    }
    DumpList->DeleteAll();
    AggregationInfo->DisposeResult(TimeStep);
    MarkForRedraw();
    return TRUE;
}

/* =================================================================================== */

bool
TMapView::HandleButtonPress(int x, int y, int ButtonState)

{
    switch (ButtonState)
    {
    case 1:

        if (MarkActive)
        {
            SecondMark.x() = X2World(x);
            SecondMark.y() = Y2World(y);
            MarkActive = FALSE;
            if (!MapViewer->OpenView(TRectangle(FirstMark, SecondMark)))
            {
                Error("HandleButtonPress: cannot create new view");
                return FALSE;
            }
        }
        else
        {
            FirstMark.x() = X2World(x);
            FirstMark.y() = Y2World(y);     
            MarkActive = TRUE;
        }
        MarkForRedisplay();
        break;
        
    default:
        return ParentClass::HandleButtonPress(x, y, ButtonState);
    }
    return TRUE;
}

/* =================================================================================== */

bool
TMapView::HandleKeyPress(int KeyCode)

{
    switch(KeyCode)
    {
    case '2':
        return ModifyFrame(0.0, -0.25, 1.0);

    case '8':
        return ModifyFrame(0.0, 0.25, 1.0);

    case '4':
        return ModifyFrame(-0.25, 0.0, 1.0);

    case '6':
        return ModifyFrame(0.25, 0.0, 1.0);

    case '-':
        return ModifyFrame(0.0, 0.0, sqrt(2.0));

    case '+':
        return ModifyFrame(0.0, 0.0, 1.0/sqrt(2.0));

    case 'm':
    case 'M':
        
        if (!SelectMode(LineViewMode))
        {
            Error("cannot switch to LineViewMode");
            return FALSE;
        }
        break;

    case 'v':
    case 'V':
        
        if (!SelectMode(VelocityViewMode))
        {
            Error("cannot switch to VelocityViewMode");
            return FALSE;
        }
        break;

    case 'd':
    case 'D':
        
        if (!SelectMode(DensityViewMode))
        {
            Error("cannot switch to VelocityViewMode");
            return FALSE;
        }
        break;

    case 'o':
    case 'O':
        
        if (!SelectMode(DumpViewMode))
        {
            Error("cannot switch to DumpViewMode");
            return FALSE;
        }
        break;

    case 'q':
    case 'Q':

        MapViewer->CloseView(this);
        break;
        
    case 'y':
    case 'Y':
        
        if (!SelectMode(DelayViewMode))
        {
            Error("cannot switch to DelayViewMode");
            return FALSE;
        }
        break;

    default:

        return ParentClass::HandleKeyPress(KeyCode);
    }
    return TRUE;
}

/* =================================================================================== */

void
TMapView::LabelDump(const char *Label)

{
    *this << GrString(GetWorldX(10), GetWorldY(20), (char *) Label, 0);
}

// =================================================================================== 

bool
TMapView::ModifyFrame(double PercentX, double PercentY, double Zoom)

{
    double DeltaX;
    double DeltaY;

    DeltaX = Frame.GetDeltaX() * PercentX;
    DeltaY = Frame.GetDeltaY() * PercentY;
    
    Frame += TLocation(DeltaX, DeltaY);
    Frame *= Zoom;
    if (DumpDesc)
    {
        DumpDesc->SetFrame(Frame);
        if (!StatManager->ModifySampling(DumpDesc))
        {
            Error("ModifyFrame: cannot modify %s", (char *) DumpDesc->GetObjectName());
            return FALSE;
        }
    }
    SetXaxis(Frame.GetMinX(), Frame.GetMaxX());
    SetYaxis(Frame.GetMinY(), Frame.GetMaxY());
    FrameChanged = TRUE;
    MapViewer->MarkViews();
    return TRUE;
}

// =================================================================================== 

bool
TMapView::SelectMode(TViewMode NewMode)

{
    if (Mode != NewMode)
    {
        MarkForRedraw();
        switch(Mode)
        {
        case LineViewMode:
        case VelocityViewMode:
        case DensityViewMode:
            
            break;
            
        case DumpViewMode:
        case DelayViewMode:
            
            StopMultilaneDump();
            break;
        }
        Mode = NewMode;
        switch(Mode)
        {
        case LineViewMode:
        case VelocityViewMode:
        case DensityViewMode:
            
            break;
            
        case DumpViewMode:
        case DelayViewMode:
            
            if (!StartMultilaneDump())
            {
                Error("SelectMode: cannot start multilane dump");
                return FALSE;
            }
            break;
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TMapView::StartMultilaneDump(void)

{
    DumpDesc = new TSamplingDescription(TMultilaneDump::ClassInfo->GetID(), 
                                        SF_MERGE, Parallel->GetLocalCPNID(), 
                                        SAMPLING_ID_CA_DUMP_BASE);
    DumpDesc->SetCollectInterval(DumpInterval);
    DumpDesc->SetFrame(Frame);
    DumpInfo = new TAggregationInfo(this, DumpDesc);
    if (!StatManager->RegisterAggregation(DumpInfo))
    {
        Error("StartMultilaneDump: cannot register %s",
              (char *) DumpInfo->GetObjectName());
    }
    if (!StatManager->StartSampling(DumpDesc, TRUE))
    {
        Error("StartMultilaneDump: cannot start %s",
              (char *) DumpDesc->GetObjectName());
    }
    return TRUE;
}

// =================================================================================== 

void
TMapView::StopMultilaneDump(void)

{
    if (DumpDesc)
    {
        StatManager->StopSampling(DumpDesc);
        DeleteObject(DumpDesc);
    }
    if (DumpInfo)
    {
        StatManager->CancelAggregation(DumpInfo);
        DeleteObject(DumpInfo);
    }
}

// =================================================================================== 
#undef ParentClass
