/*      file     : common/TBaseObject.h
        project  : General Toolbox
        author   : Marcus Rickert
        RCS      : $Id: TBaseObject.h,v 1.11 2005/03/25 11:04:59 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

#ifndef TBASEOBJECT_H
#define TBASEOBJECT_H

// ***INCLUDES************************************************************************

#include <stdarg.h>

#include <Basic.h>

// ***DEFINES*************************************************************************

#define BO_FLAG_FAILED           0x00000001
#define BO_FLAG_FIRST_FREE_FLAG (BO_FLAG_FAILED << 1)

#define Fail(Message) \
{ \
      Error("CONTRUCTOR: %s", Message); \
      _Fail(); \
      return; \
}

// ***CLASSES*************************************************************************

class TString;   /* forward */

/** 
  Class that all other classes should be derived from.
  
  This class provides basic virtual functions for comparison and printing. It also
  defines a name for each class and a description for a given instance of a class.

  */

class TBaseObject

{
 public:

    inline TBaseObject(long theFlags = 0);
    virtual ~TBaseObject(void) { };

    inline bool
    AreBaseFlagsSet(long Mask) const;

    /** Compare the current instance with another instance of the same class.

     This method must be overwritten, if the class is to be used in sorted containers.

     @param OtherObject pointer to the second innstance which may have to be type-casted.
     @param CompareCode integer that selects the sorting mode, defaults to 0
     @return <B>-1</B> if the local instance is 'less' than the other instance,
     <B>0</B> if the instances are equivalent,
     <B>1</B> if the local instance id 'greater' than the other instance
     */
    virtual int
        Compare (TBaseObject *OtherObject, int CompareCode = 0) const;

    /** Create an identical copy of the instance. 
     This method must be overwritten to provide correct handling of containers. */

    virtual TBaseObject *
        Copy(void) const;

    /** Print an error to the active output device. 
     The message will be prefixed by an 'ERROR' string. 
     
     @param Format format string equivalent to the one defined in printf(3).
     */
    virtual void
        Error(const char *Format ...) const;

    inline void
        _Fail(void);
    /** Test if the instantiation of the class failed.

     If an instance was created using dynamic allocation, this method can be used to 
     test whether the instantiation was successfull.

     @return TRUE if instantiation failed.
     */
    inline bool
        Failed(void) const;


    /** Return a unique class name.

     This method must be overwritten to provide a unique class name. This name will
     used for all messages printed. */
    virtual TString
        GetClassName(void) const;
    /** */
    inline long 
        GetBaseFlags(void) const;
    /** */
    virtual TString
        GetInstanceName(void) const;
    /** */
    virtual TBaseObject *
        GetInstancePtr(void);
    /** */
    virtual TString
        GetObjectName(void) const;
    /** */
    virtual void
        NotOverloaded(const char *MethodName) const;
#ifdef TDEBUG
    /** */
    virtual void
        Print(unsigned long Mask, const char *Format ...) const;
#endif
    /** */
    virtual void
        Print(const char *Format ...) const;
    /** */
    inline void
    ResetBaseFlags(long Mask);
    /** */
    inline void
    SetBaseFlags(long Mask);
    /** */
    virtual void
        Warning(const char *Format ...) const;

 protected:
    
    long Flags;
};

// ***PROTOTYPES********************************************************************** 
#ifdef __cplusplus
extern "C" {
#endif

    inline int
        CompareDouble(double D1, double D2);
    inline int
        CompareLong(long L1, long L2);

#ifdef __cplusplus
}
#endif
// ***INLINE METHODS******************************************************************

#ifndef NO_INLINE
#  include "TBaseObject.i.C"
#endif

// ===================================================================================
// Local Variables:
// mode:c++
// End:
#endif
