/*      file     : common/TConfiguration.C
        project  : General Toolbox
        author   : Marcus Rickert
        RCS      : $Id: TConfiguration.C,v 1.13 2005/05/01 14:57:58 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <cassert>

#include "LexYacc.h"

#include "CommonConfig.h"
#include "TDebug.h"
#include "TBaseStream.h"
#include "TEnvironment.h"
#include "TConfiguration.h"

// ***LOCAL VARIABLES***************************************************************** 

TConfiguration *Configuration = 0;

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include "TConfiguration.i.C"
#endif

// =================================================================================== 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TConfigEntry::TConfigEntry(TBaseStream &aStream)

{
    Decode(aStream);
}

// =================================================================================== 

TConfigEntry::TConfigEntry(const char *aGroupname, const char *aName, const char *aValue)

{
    assert(aGroupname != 0);
    assert(aName != 0);
    assert(aValue != 0);

    Groupname = aGroupname;
    Name = aName;
    Default = aName;
    Value = aValue;
}

// =================================================================================== 

TConfigEntry::TConfigEntry(const char *aGroupname, const char *aName, long aValue)

{
    assert(aGroupname != 0);
    assert(aName != 0);

    char StringValue[32];

    Groupname = aGroupname;
    Name = aName;
    sprintf(StringValue, "%ld", aValue);
    Value = StringValue;
}

// =================================================================================== 

TConfigEntry::TConfigEntry(const char *aGroupname, const char *aName, double aValue)

{
    assert(aGroupname != 0);
    assert(aName != 0);

    char StringValue[32];

    Groupname = aGroupname;
    Name = aName;
    sprintf(StringValue, "%f", aValue);
    Value = StringValue;
}

// =================================================================================== 


TConfigEntry::TConfigEntry(const TConfigEntry &anEntry)

{
    assert(anEntry.Groupname.GetPtr() != 0);
    assert(anEntry.Name.GetPtr() != 0);
    assert(anEntry.Value.GetPtr() != 0);

    Groupname = anEntry.Groupname;
    Name = anEntry.Name;
    Value = anEntry.Value;
}

// =================================================================================== 

bool
TConfigEntry::Decode(TBaseStream &aStream)

{
    int result = Groupname.Decode(aStream) AND
        Name.Decode(aStream) AND
        Value.Decode(aStream);

    assert(Groupname.GetPtr() != 0);
    assert(Name.GetPtr() != 0);
    assert(Value.GetPtr() != 0);

    return result;
}

// =================================================================================== 

bool
TConfigEntry::Encode(TBaseStream &aStream) const

{
    return (Groupname.Encode(aStream) AND
            Name.Encode(aStream) AND
            Value.Encode(aStream));
}

// =================================================================================== 

TString
TConfigEntry::GetClassName(void) const

{
    return "TConfigEntry";
}

// =================================================================================== 

TString
TConfigEntry::GetInstanceName(void) const

{
    assert(Groupname.GetPtr() != 0);
    assert(Name.GetPtr() != 0);
    assert(Value.GetPtr() != 0);

    char Line[1024];

    sprintf(Line, "%s.%s Value='%s'", Groupname(), Name(), Value());
    return Line;
}

// =================================================================================== 

bool
TConfigEntry::WriteToFile(FILE *output)

{
    fprintf(output, "# %s.%s='%s'\n", Groupname(), Name(), Value());
    return TRUE;
}

// =================================================================================== 

TConfiguration::TConfiguration(void) :

Table(CM_AUTO_DISPOSE),
PreviousEntry("<UNKNOWN>", "<UNKNOWN>")

{
    if (Configuration != 0)
    {
        Fail("Configuration already active");
    }
    Configuration = this;
}

// =================================================================================== 

TConfiguration::~TConfiguration(void)

{
    Table.DeleteAll();
    Configuration = 0;
}

// =================================================================================== 

bool
TConfiguration::Decode(TBaseStream &aStream)

{
    int Count;
    int i; 

    if (!aStream.Decode(Count))
        return FALSE;
    for (i = 0; i < Count; i++)
    {
        TConfigEntry Entry(aStream);
        
        Set(Entry);
    }
    debug(Print(DM_CONFIG, "decoded %d configuration entries", Count));
    return TRUE;
}

// =================================================================================== 

bool
TConfiguration::Encode(TBaseStream &aStream) const

{
    TEnumerator< TConfigEntry > CEntry(&Table);
    long Count;

    Count = Table.GetNrOfElements();
    if (!aStream.Encode(Count))
        return FALSE;
    CEntry.Enumerate();
    while (CEntry())
    {
        if (!CEntry->Encode(aStream))
            return FALSE;
        CEntry++;
    }
    return TRUE;
}

// =================================================================================== 

TString
TConfiguration::GetClassName(void) const

{
    return "TConfiguration";
}

// =================================================================================== 

const char *
TConfiguration::GetString(const char *Groupname, const char *Name)

{
    TStringContainerEntry< TConfigEntry > *ContainerEntry;

    if (Groupname != 0 && Name != 0)
        PreviousEntry = TConfigEntry(Groupname, Name);
    if ((ContainerEntry = Table.Find((char*) PreviousEntry.GetKey(), CONFIG_TABLE_STRCMP)))
        return ((TConfigEntry *) *ContainerEntry)->GetValue();
    else
    {
        Warning("no config entry for %s found.", (char *) PreviousEntry.GetKey());
        return "<INVALID>";
    }
}

// =================================================================================== 

bool
TConfiguration::OptionActive(const char *Option)

{
    return Exists(OPTION_GROUP, Option);
}

// =================================================================================== 

int
TConfiguration::ParseError(const char *ErrorString, int LineNumber)

{
    Error("Parse Error: '%s' in line %d", ErrorString, LineNumber);
    return FALSE;
}

// =================================================================================== 

bool
TConfiguration::ReadFile(const char *Filename)

{
    FILE *Input;
    int   Result;

    if ((Input = fopen(Filename, "r")) == 0)
    {
        //! \todo check ioresult
        Warning("cannot open config file '%s'", Filename);
        return FALSE;
    }
    Print(DM_FILE, "reading file '%s'", Filename);
    yyin = Input;
    Result = yyparse();
    fclose(Input);
    if (Result != 0)
        Warning("Error while parsing config file '%s'", Filename);
    return Result == 0;
}

// =================================================================================== 

void
TConfiguration::Set(const TConfigEntry &anEntry)

{
    TStringContainerEntry < TConfigEntry > *ContainerEntry;

    assert(((void *) anEntry.GetKey()) != 0);
    assert(anEntry.GetKey().GetPtr() != 0);
    assert(anEntry.GetValue() != 0);

    if ((ContainerEntry = Table.Find((char *) anEntry.GetKey(), CONFIG_TABLE_STRCMP)))
    {
        TConfigEntry *Entry = *ContainerEntry;

        Print(DM_CONFIG, "fill existing entry '%s' with new value '%s'", 
              (char *) Entry->GetInstanceName(), anEntry.GetValue());
        Entry->SetValue(anEntry.GetValue());
    }
    else
    {
        TConfigEntry *Entry = new TConfigEntry(anEntry);

        Print(DM_CONFIG, "new entry '%s'", (char *) Entry->GetInstanceName());
        Table.Insert(new TStringContainerEntry< TConfigEntry > 
                     ((char *) anEntry.GetKey(), Entry), TRUE, CONFIG_TABLE_STRCMP);
    }
}
    
// =================================================================================== 

void
TConfiguration::SetDefault(const TConfigEntry &anEntry)

{
    TStringContainerEntry < TConfigEntry > *ContainerEntry;

    if ((ContainerEntry = Table.Find((char *) anEntry.GetKey(), CONFIG_TABLE_STRCMP)))
    {
        TConfigEntry *Entry = *ContainerEntry;

        debug(Print(DM_CONFIG, "fill %s with new value '%s'", 
                    (char *) Entry->GetInstanceName(), anEntry.GetValue()));
        Entry->SetDefault(anEntry.GetValue());
    }
    else
    {
        TConfigEntry *Entry = new TConfigEntry(anEntry);

        Table.Insert(new TStringContainerEntry< TConfigEntry > 
                     ((char *) anEntry.GetKey(), Entry), TRUE, CONFIG_TABLE_STRCMP);
        debug(Print(DM_CONFIG, "new entry %s", (char *) Entry->GetInstanceName()));
    }
}
    
// =================================================================================== 

bool
TConfiguration::WriteToFile(FILE *output)

{
    TEnumerator< TConfigEntry > CEntry(Table);

    CEntry.Enumerate();
    while (CEntry())
    {
        //      TConfigEntry *CEntry2 = CEntry;

        if (!CEntry->WriteToFile(output))
            return FALSE;
        CEntry++;
    }
    fprintf(output, "#\n");
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
