/*      file     : common/TEnvironment.C
        project  : General Toolbox
        author   : Marcus Rickert
        RCS      : $Id: TEnvironment.C,v 1.27 2005/05/01 14:57:58 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#if defined (LINUX) || defined(MPI_LINUX)
#  include <getopt.h>
#elif defined(SUN4SOL2)
#else

extern "C" int getopt (int, char * const*, const char *);

#endif

#include "CommonConfig.h"

#include "File.h"
#include "Time.h"
#include "TTimer.h"
#include "TDebug.h"
#include "TRandom.h"
#include "TBaseStream.h"
#ifdef OLD_CONFIG
#include "TConfig.h"
#else
#include "TConfiguration.h"
#endif
#include "TConfigHandler.h"
#include "TEnvironment.h"

// ***DEFINES************************************************************************* 

#define DEFAULT_CONFIG_NAME "Config"
#define DEFAULT_DEBUG_MASK  0

#define DEFAULT_APP_CONFIG_FORMAT   "%%X/%s"
#define DEFAULT_LOCAL_CONFIG_FORMAT "%%W/%s"

#define ENV_PRINT_PREFIX    ""
#define ENV_ARCHITECTURE    "UNKNOWN_ARCH"
#define ENV_PROCESS_NAME    "UNKNOWN_PROCESS_NAME"
#define ENV_RUN_NAME        "UNKNOWN_RUN_NAME"
#define ENV_HOST_NAME       "UNKNOWN_HOST_NAME"
#define ENV_OPTION_STRING   "a:b:c:d:e:f:g:h:i:j:k:l:m:n:o:p:q:r:s:t:u:v:w:x:y:z:" \
                            "A:B:C:D:E:F:G:H:I:J:K:L:M:N:O:P:Q:R:S:T:U:V:W:X:Y:Z:"
#define ENV_APP_BASE_DIR    "TOOLBOX_APP_HOME"
#define ENV_DEFAULT_DIR     "."
#define ENV_APP_REL_PATH    "application"

// d:<debugmask>

// ***GLOBAL VARIABLES****************************************************************  

extern "C" 
{
    TEnvironment  *Environment = 0;
    char          *_PrintPrefix = "";
    unsigned long  _DebugMask = 0;
}

// ***FUNCTIONS*********************************************************************** 


void
_PrintString(const char *String)

{
    fprintf(stderr, "%s%s\n", _PrintPrefix, String);
}

// ===================================================================================

void
Print(const char *Format, ...)

{
    va_list argptr;
    char    line[2048];

    va_start(argptr, Format);
    vsprintf(line, Format, argptr);
    _PrintString(line);
    va_end(argptr);
}


// ===================================================================================

void
PrintDebug(unsigned long DebugFlag, const char *Format, ...)

{
    va_list argptr;
    char    line[2048];

    if (DebugFlag & _DebugMask)
    {
        va_start(argptr, Format);
        vsprintf(line, Format, argptr);
        _PrintString(line);
        va_end(argptr);
    }
}


// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#include "TEnvironment.i.C"
#endif

// =================================================================================== 

void
Initialize(void)

{
    if (Environment != 0)
        {
            _PrintPrefix = "";
            Environment = 0;
            Print("WARNING: global variable(s) not initialized!");
        }
}

// =================================================================================== 
#define ParentClass
// =================================================================================== 

TEnvironment::TEnvironment(int sysargc, char **sysargv)

{
    int i;

    my_argc = sysargc;
    my_argv = new char *[my_argc+1];

    for (i = 0; i <= my_argc; i++)
    {
        my_argv[i] = sysargv[i];
        // Print("argv[%d]='%s'", i, my_argv[i]);
    }

    _Init();
    if (Environment)
    {
        Fail("another instance of TEnvironment already active");
    }
    Environment = this;
#ifdef OLD_CONFIG
    Config = new TConfig;
#else
    Configuration = new TConfiguration;
    Configuration->Set(TConfigEntry(ENV_GROUP, ENV_NAME_DEBUGMASK, (long) DEFAULT_DEBUG_MASK));
#endif
    debug(Print(DM_LIFESPAN, "Hello World"));
}

// =================================================================================== 

TEnvironment::~TEnvironment(void)

{
    Deactivate();
    debug(Print(DM_LIFESPAN, "Goodbye"));
}

// =================================================================================== 

bool
TEnvironment::Activate(void)

{
    char     Filename[512];
    TString  ApplicationFilename;
    char    *PWD;

    GetStrings();
    sprintf(Filename, DEFAULT_APP_CONFIG_FORMAT, DEFAULT_CONFIG_NAME);
#ifdef OLD_CONFIG
    if (!Config->Load(GetApplicationFilename(Filename)))
    {
        Error("while reading configuration file");
        return FALSE;
    }
    if (!ParseOptions())
    {
        Error("while parsing command line options");
        return FALSE;
    }
#else
    if (!ParseOptions())
    {
        Error("while parsing command line options");
        return FALSE;
    }
    if ((PWD = getenv("PWD"))) {
        Configuration->Set(TConfigEntry(ENV_GROUP, ENV_NAME_WORKDIR, PWD));
        WorkDir = PWD;
    }
    else
        Warning("cannot retrieve environment setting PWD");
    ApplicationFilename = GetApplicationFilename(Filename);
    if (!Configuration->ReadFile(ApplicationFilename)) {
        Warning("could not read application configuration '%s' -> reverting to local config...", 
                (char *) ApplicationFilename);
        if (!LoadLocalConfig())
            return FALSE;
    }
#endif
    InterpretConfig();
    return TRUE;
}

// =================================================================================== 

bool
TEnvironment::Deactivate(void)

{
    OverallTimer->Stop();
    Print(DM_TIME, "overall running time: %.1f [s] accounted for %.1f [s]", 
          Timer->GetElapsedTime(), OverallTimer->GetElapsedTime());
    DeleteObject(Timer);
    DeleteObject(OverallTimer);
    DeleteObjectArray(my_argv);
#ifdef OLD_CONFIG
    DeleteObject(Config);
#else
    DeleteObject(Configuration);
#endif
    Environment = 0;    
    return TRUE;
}

// =================================================================================== 

bool
TEnvironment::DecodeConfiguration(TBaseStream &Stream)

{
    long _DebugMask;

    _DebugMask = Debug.GetMask();
    if (!Stream.Decode(_DebugMask))
        return FALSE;
    Debug.SetMask(_DebugMask);
    if (!Configuration->Decode(Stream))
        return FALSE;
    InterpretConfig();
    return TRUE;
}

// =================================================================================== 

bool
TEnvironment::EncodeConfiguration(TBaseStream &Stream)

{
    long _DebugMask;

    _DebugMask = Debug.GetMask();
    return (Stream.Encode(_DebugMask) AND
            Configuration->Encode(Stream));
}

// =================================================================================== 

TString
TEnvironment::GetClassName(void) const

{
    return "TEnvironment";
}

// =================================================================================== 

TString
TEnvironment::GetFilename(const char *BaseDir, const char *Path) const

{
    TString RelativePath;
    TString Filename;
    TString Result;

    RelativePath = ReplacePathTokens(GetPathPart(Path));
    Filename = ReplacePathTokens(GetFilenamePart(Path));
    Result = CombinePaths(BaseDir, RelativePath);
    return CombinePaths(Result, Filename);
}

/* =================================================================================== */

bool
TEnvironment::GetStrings(void)

{
    const char *Home;
    const char *BaseDir;

    Executable = GetFilenamePart(my_argv[0]);
    if ((Home = getenv("HOME")))        
        HomeDir = Home;
    else
        Warning("no HOME environment found, using '%s'", 
                (char *) HomeDir);
    if ((BaseDir = getenv(ENV_APP_BASE_DIR)))
        AppBaseDir = BaseDir;
    else
    {
        AppBaseDir = CombinePaths(HomeDir, ENV_APP_REL_PATH);
        Warning("no %s environment found, using '%s'", 
                ENV_APP_BASE_DIR, AppBaseDir());
    }
    return TRUE;
}

// =================================================================================== 

void
TEnvironment::_Init(void)

{
    Timer = new TTimer;
    Timer->Start();
    OverallTimer = new TTimer;
    OverallTimer->Start();
    PrintPrefix = ENV_PRINT_PREFIX;
    _PrintPrefix = PrintPrefix.GetPtr();
    Architecture = ENV_ARCHITECTURE;
    ProcessName = ENV_PROCESS_NAME;
    HostName = ENV_HOST_NAME;
    HomeDir = ENV_DEFAULT_DIR;
    WorkDir = ENV_DEFAULT_DIR;
    AppBaseDir = ENV_DEFAULT_DIR;
    RunName = ENV_RUN_NAME;
    LocalID = DEFAULT_LOCAL_ID;
    OptionString = ENV_OPTION_STRING;
    Config = 0;
    ThrowCore = FALSE;
    CheckLocalConfig = TRUE;
    _DebugMask = 0; // DM_MSG | DM_MSG_CONTROL | DM_PARALLEL;    
}

/* =================================================================================== */

bool
TEnvironment::InterpretOption(int c, char *OptArg, bool &Handled)

{
    char OptionName[2];

    Handled = FALSE;
    switch(c)
    {
    case 'd':
        Handled = TRUE;
        _DebugMask = atoi(OptArg);
        break;

    default:
        
#ifdef OLD_CONFIG
        if (!Config->CommandLineConfigItem(OptArg))
        {
            Error("error while handling option '%c'", c);
            return FALSE;
        }
#else
        OptionName[0] = c;
        OptionName[1] = 0;
        Configuration->Set(TConfigEntry(OPTION_GROUP, OptionName, OptArg));
        Handled = TRUE;
#endif
        break;
    }
    return TRUE;
}

/* =================================================================================== */

bool
TEnvironment::LoadLocalConfig(void) 

{
    char    FilenamePattern[512];
    TString Filename;

    sprintf(FilenamePattern, DEFAULT_LOCAL_CONFIG_FORMAT, DEFAULT_CONFIG_NAME);
    Filename = GetApplicationFilename(FilenamePattern);
    if (!fexist(Filename))
    {
        Warning("local configuration file '%s' does not exist!", (char *) Filename);
        return TRUE;
    }
    if (!Configuration->ReadFile(Filename)) {
        Warning("error while reading local configuration file '%s'", (char *) Filename);
        return FALSE;
    }
    InterpretConfig();
    return TRUE;
}

/* =================================================================================== */
    
bool
TEnvironment::ParseOptions(void)

{
  int  c;
  int  ParameterNr;
  bool Handled;

  extern char *optarg;
  extern int   opterr;
  extern int   optind;

  opterr = 0; /* prevent errors */
  do
  {
      switch(c=getopt(my_argc, my_argv, OptionString()))
      {
      case EOF:
          break;
            
      case '?':
      case ':':
          Warning("ignoring option '%c'", c);
          break;
          
      default:
          Handled = FALSE;
          if (!InterpretOption(c, optarg, Handled))
          {
              Error("error while handling option '%c'", c);
              return FALSE;
          }
          if (!Handled)
              Warning("option '%c' not handled", c);
      }
  }
  while (c != EOF);
  ParameterNr = 1;
  while (optind < my_argc)
  {
      if (!InterpretOption('0'+ParameterNr, my_argv[optind], Handled))
          return FALSE;
      if (!Handled)
          Warning("%d. parameter '%s' not handled", ParameterNr, my_argv[optind]);
      ParameterNr++;
      optind++;
  }
  return TRUE;
}

// ===================================================================================

void
TEnvironment::PrintString(const char *Format, ...) const

{
    va_list argptr;
    char    line[2048];

    va_start(argptr, Format);
    vsprintf(line, Format, argptr);
    _PrintString(line);
    va_end(argptr);
}

/* =================================================================================== */

TString
TEnvironment::ReplacePathTokens(const char *Filename) const

{
    int         Source, Dest;
    int         Length;
    int         i;
    char        Result[1024];
    const char *Replacement = 0;
    char        Value[16];

    Source = 0;
    Dest = 0;
    while (Filename[Source])
    {
        switch(Filename[Source])
        {
        case '%':
            Source++;
            switch (Filename[Source])
            {
            case 0:
                Error("cannot replace single %% in filename '%s'", Filename);
                return "ERROR";

            case 'X':
                Replacement = Executable;
                break;

            case 'A':
                Replacement = Architecture;
                break;

            case 'D':
                sprintf(Value, "%ld", (long) getpid());
                Replacement = Value;
                break;

            case 'W':
                Replacement = WorkDir;
                break;

            case 'H':
                Replacement = HostName;
                break;

            case 'h':
                Replacement = HomeDir;
                break;

            case 'P':
                Replacement = ProcessName;
                break;

            case 'R':
                Replacement = RunName;
                break;
                
            default:
                Warning("unknown path character '%c'", Filename[Source]);
                Replacement = "ERROR";
            }
            i = 0;
            Length = 0;
            while (Replacement[i])
            {
                if (Replacement[i] != ' ')
                    Result[Dest+Length++] = Replacement[i];
                i++;
            }
            Result[Dest+Length] = 0;
            break;

        default:
            Result[Dest] = Filename[Source];
            Length = 1;
        }
        Dest += Length;
        Source ++;
    }
    Result[Dest] = 0;
    return Result;
}

// ===================================================================================

void
TEnvironment::InterpretConfig(void)

{
    if (Configuration->Exists(ENV_GROUP, ENV_NAME_WORKDIR))
    {
        SetWorkDir(Configuration->GetString(ENV_GROUP, ENV_NAME_WORKDIR));
        Print(DM_CONTROL, "WorkDir set to '%s'", (char *) WorkDir);
        CheckLocalConfig = TRUE;
    }
    if (Configuration->Exists(ENV_GROUP, ENV_NAME_SEED))
    {
        long BaseSeed = Configuration->GetLong();
        
        if (BaseSeed == -1)
        {
            struct timeval CurrentTime;

            GetTime(&CurrentTime);
            BaseSeed = CurrentTime.tv_usec;
        }
        TRandom::BaseSeed = BaseSeed;
    }
    _DebugMask = Configuration->GetLong(ENV_GROUP, ENV_NAME_DEBUGMASK);
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGLIFESPAN)? DM_LIFESPAN:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGCONTROL)? DM_CONTROL:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGCODE)? DM_CODE:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGEVENT)? DM_EVENT:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGMSG)? DM_MSG:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGMSGCONTROL)? DM_MSG_CONTROL:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGPARSE)? DM_PARSE:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGPARALLEL)? DM_PARALLEL:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGSTRING)? DM_STRING:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGABTREE)? DM_AB_TREE:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGSIGNAL)? DM_SIGNAL:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGALARM)? DM_ALARM:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGLEX)? DM_LEX:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGYACC)? DM_YACC:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGCONFIG)? DM_CONFIG:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGCLASS)? DM_CLASS:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGBOUNDARY)? DM_BOUNDARY:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGSTATSAMPLE)? DM_STAT_SAMPLE:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGSTATAGGR)? DM_STAT_AGGR:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGTIMESTEP)? DM_TIME_STEP:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGMIGRATION)? DM_MIGRATION:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGTIME)? DM_TIME:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGLOAD)? DM_LOAD:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGCONTAINER)? DM_CONTAINER:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGOBJECT)? DM_OBJECT:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGFILE)? DM_FILE:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGBALANCE)? DM_BALANCE:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGCOLLISION)? DM_COLLISION:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGTOPOLOGY)? DM_TOPOLOGY:0;
    _DebugMask |= Configuration->Exists(ENV_GROUP, ENV_NAME_DEBUGROUTE)? DM_ROUTE:0;
    Debug.SetMask(_DebugMask);    
}

// ===================================================================================

bool
TEnvironment::Run(void)

{
    if (CheckLocalConfig)
    {
        if (!LoadLocalConfig())
            return FALSE;
    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
