/*      file     : parallel/TCommunication.h
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TCommunication.h,v 1.17 2005/03/25 11:06:20 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

#ifndef TCOMMUNICATION_H
#define TCOMMUNICATION_H

// ***INCLUDES************************************************************************ 

#include <TBaseObject.h>
#include <TDLList.h>
#include "ParallelTypes.h"
#include "TDestination.h"
#include "TBaseMessage.h"

// ***FORWARDS************************************************************************

class TBaseMessage;
class TCommunicationControl;
class TCPNInfo;
class TMessageRecipient;
class TTimer;

// ***DEFINES************************************************************************* 

#define MAX_CONTROLLERS    2
#define MAX_RECIPIENTS    16
#define MAX_PREFETCH       8

#define TAG_MASK_ALL      0x7FFFFFFF

#define MSG_FLAG_DEFAULT  (MSG_FLAG_EVENT | MSG_FLAG_AUTO | MSG_FLAG_DELETE)

// ***CLASSES************************************************************************* 

/** Class handling messages.
  All messages are sent and received through using this class. */

class TCommunication : public TBaseObject

{
 public:

    TCommunication(void);

    virtual ~TCommunication(void);

    /** Activate the communication class. This method must me called before a message 
      can be sent. */
    bool
        Activate(void);

    /** Activates a subset of message tags.
      Only messages containing tags that have been activated through this method will be
      processed.
      @param ActivateMask bit mask of tags to be activated.
      */
    void
        ActivateTags(TMessageTag ActivateMask);

    /** Cancel a message recipient. */

    bool
    CancelRecipient(TMessageRecipient *Recipient);

    /** Check capabilities of communication subsystem. 
      @return TRUE if all required capabilities are available.
      @param RequiredCapabilities bit mask of required capabilities. The masks are defined
      in file ParallelTypes.h      
     */
    bool
    CheckCapabilities(long RequiredCapabilities) const;

    /** Send a message.
      Sends a message which had been created using the GetNewMessage method. */
    bool
        Commit(TBaseMessage *&aMessage);

    /** Deactivates a subset of message tags.
      Only messages containing tags that have been activated through this method will be
      processed.
      @param DeactivateMask bit mask of tags to be deactivated.
      */
    void
        DeactivateTags(TMessageTag DeactivateMask);

    /** Dispose of a message that was received through #ReceiveMessage#. Be sure to
      call this method always. Otherwise some message buffers may never be deallocated! */
    bool
        DestroyMessage(TBaseMessage *&aMessage) const;

    bool
        Enroll(void);
    /** Creates and returns a new message for sending.
      @return TRUE if successfull
      @param aMessage reference to pointer that will hold the message. This message will
      have to sent using the Commit method. */

    TTimer *
         GetIdleTimer(void);


    bool
        GetNewMessage(TBaseMessage *&aMessage, TDestination theDestination, 
                      TMessageTag theTag, TMessageType theType, 
                      TMessageFlags theFlags = MSG_FLAG_DEFAULT);
    /** Wait for incoming messages.
      This method will wait for incoming messages and pass them to the registered message
      recipients. 
      @return TRUE, if successful
      @param Block if TRUE, return only if message has been received. */

    bool
        MessageLoop(bool Block = TRUE, TCPNID = SOURCE_ALL);

    /** Process all pending messages. */

    bool
    ProcessPendingMessages(void);

    /** Receive message explicitly. This method wait for a message matching the given 
     specifications. Note that all other (non-matching) messages will be queue which may
     result in an overflow. If this happens, you may want to in the setting of [PVM].MaxBuffer. */
    bool
        ReceiveMessage(TBaseMessage *&Message, TDestination theSender,
                       TMessageTag theTag, TMessageType theType);
    
    /** Register a class instance that will receive messages. 
      Each class instance that will receive messages, has to be registered using this method.
      @param theTagMask bit mask defining which message tags should be passed to this class.
      @param theRecipient pointer to the class instance that will receive the messages.
     */
    bool
        RegisterRecipient(TMessageTag theTagMask, TMessageRecipient *theRecipient);

    /** Short cut to send a message consisting of single event. There is no additional data
      encoded. */
    bool
        SendEvent(TEventID EventID, TDestination Destination, TMessageTag Tag, 
                  TMessageType Type, TMessageFlags Flags = MSG_FLAG_DEFAULT);

#include "Reduce.Headers.h"

 private:

    bool
        CheckReduceParameters(TBaseOperation Operation, TDestination &Destination,
                              TCPNID &Root, TCommunicationControl *&Control);
    bool
        Deactivate(void);
    bool
        DeliverMessage(TBaseMessage *aMessage, bool &ExitLoop, bool &Success);
    void
        DumpPendingMessages(void);
    virtual TString
        GetClassName(void) const;
    virtual TString
        GetInstanceName(void) const;
    bool
        GetControl(long Capabilities, TCommunicationControl *&Control);
    bool
        FillMessageList(bool Block);
    void
        _Init(void);

    TCPNID      LocalCPNID;
    TMessageID  NextMessageID;

    int                    NrOfControllers;
    TCommunicationControl *Controllers[MAX_CONTROLLERS];

    TDLList<TBaseMessage> *PendingMessages;
    int                    NrOfRecipients;
    TMessageRecipient     *Recipients[MAX_RECIPIENTS];
    TMessageTag            TagMasks[MAX_RECIPIENTS];
    TMessageTag            ActiveTagMask;
    bool                   TagMaskHasChanged;
    TTimer                *CommunicationTimer;
    TTimer                *ReceiveTimer;
    TTimer                *SendTimer;
    TTimer                *IdleTimer;
    long                   MessageTimeOut;
};

// ***EXTERNALS*********************************************************************** 

extern TCommunication *Communication;

// ===================================================================================
// Local Variables:
// mode:c++
// End:
#endif
