/*      file     : parallel/TParallelControl.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TParallelControl.C,v 1.27 2005/03/25 11:06:20 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

extern "C"
{
#include <math.h>
}

#include <TString.h>
#include <TEnvironment.h>
#include <TConfiguration.h>

#include "ParallelTypes.h"
#include "ParallelConfig.h"
#include "TCommunication.h"
#include "TCPNInfo.h"
#include "THostInfo.h"
#include "TBaseMessage.h"
#include "TEventHandler.h"
#include "TGridTopology.h"
#include "TBusTopology.h"
#include "TParallelControl.h"

// ***DEFINES************************************************************************* 

#define DEFAULT_USEGRID   0L
#define DEFAULT_GRIDCOLS -1L
#define DEFAULT_GRIDROWS -1L

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include "TParallelControl.i.C"
#endif

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TParallelControl::TParallelControl(void)

{
    _Init();
}

/* =================================================================================== */

TParallelControl::~TParallelControl(void)

{
    if (Mutable)
    {
        if (Mutable->CPNInfoList)
        {
            Mutable->CPNInfoList->DeleteAll(TRUE);
            DeleteObject(Mutable->CPNInfoList);
        }
        if (Mutable->HostInfoList)
        {
            Mutable->HostInfoList->DeleteAll(TRUE);
            DeleteObject(Mutable->HostInfoList);
        }
        DeleteObject(Mutable);
    }
    DeleteObject(EventHandler);
}

// =================================================================================== 

bool
TParallelControl::Activate(void)

{
    return TRUE;
}

// =================================================================================== 

bool
TParallelControl::AddCPN(THostInfo *, int &)

{
    Warning("AddCPN: ignoring request");
    return TRUE;
}

// ===================================================================================

bool
TParallelControl::AddCPNInfo(const TCPNInfo *aCPNInfo)

{
    debug(Print(DM_PARALLEL, "adding CPN info %s", 
                (char *) aCPNInfo->GetInstanceName()));
    Mutable->CPNInfoList->Append(aCPNInfo);
    return TRUE;
}

// ===================================================================================

void
TParallelControl::DetermineBestGrid(int &Rows, int &Cols, int Count, int &Remainder) const

{
    if (Rows == DEFAULT_GRIDROWS AND Cols != DEFAULT_GRIDROWS)
    {
        Rows = Count / Cols;
    }
    else
        if (Rows != DEFAULT_GRIDROWS AND Cols == DEFAULT_GRIDROWS)
        {
            Cols = Count / Rows;
        }
        else
            if (Rows == DEFAULT_GRIDROWS AND Cols == DEFAULT_GRIDROWS)
            {
                Rows = (int) sqrt(Count);
                while (Cols == DEFAULT_GRIDCOLS)
                    if (Count % Rows == 0)
                        Cols = Count / Rows;
                    else
                        Rows--;
            }
    Remainder = Count - (Rows * Cols);
}

// ===================================================================================

bool
TParallelControl::DeterminePhysicalTopology(void) const

{
    if (Mutable->PhysicalTopology == 0)
    {
        int GridCols, GridRows;

        TConfiguration *C = Environment->GetConfiguration();

        if (C->GetLong(PARALLEL_GROUP, PARALLEL_NAME_USEGRID))
        {
            int Remainder;

            GridRows = (int) C->GetLong(PARALLEL_GROUP, PARALLEL_NAME_GRIDROWS);
            GridCols = (int) C->GetLong(PARALLEL_GROUP, PARALLEL_NAME_GRIDCOLS);
            DetermineBestGrid(GridRows, GridCols, GetCurrentNrOfCPN(), Remainder);
            if (Remainder)
            {
                Error("cannot find a matching grid for (%dx%d) to fit on %d CPN",
                      GridRows, GridCols, GetCurrentNrOfCPN());
                return FALSE;
            }       
            Mutable->PhysicalTopology = new TGridTopology(GridRows, GridCols);
        }
        else
            Mutable->PhysicalTopology = new TBusTopology();
    }
    return TRUE;
}

// ===================================================================================

bool
TParallelControl::Disenroll(void)

{
    Enrolled = FALSE;
    return TRUE;
}

// =================================================================================== 

bool
TParallelControl::Enroll(void)

{
    Enrolled = TRUE;
    Communication->RegisterRecipient(MTAG_PARALLEL_CONTROL, this);
    return TRUE;
}

// =================================================================================== 

TCPNInfo *
TParallelControl::GetCPNInfo(THostInfo *aHostInfo, TCPNID theCPNID) const

{
    char      Name[512];
    TCPNInfo *CPNInfo;

    if ((CPNInfo = new TCPNInfo) == 0)
    {
        Error("GetCPNInfo: out of memory");
        return 0;
    }
    CPNInfo->ID = theCPNID;
    sprintf(Name, "%s_%d", (char *) aHostInfo->GetShortName(), 
            aHostInfo->NrOfCPN++);
    CPNInfo->Name = Name;
    CPNInfo->HostID = aHostInfo->ID;
    CPNInfo->DebugMask = aHostInfo->DebugMask;
    CPNInfo->Flags |= CPNINFO_FLAG_MODIFIED;
    CPNInfo->Performance = aHostInfo->Performance;
    return CPNInfo;
}

// =================================================================================== 

TCPNInfo *
TParallelControl::GetCPNInfoByID(TCPNID SearchID) const

{
    TCPNInfoEnum CInfo;

    if (!CheckCPNInfoList())
        return 0;
    CInfo.WalksForwardThrough(Mutable->CPNInfoList);
    while (CInfo())
    {
        if (CInfo->ID == SearchID)
            return CInfo;
        CInfo++;
    }
    return 0;
}

// =================================================================================== 

const TCPNInfoList *
TParallelControl::GetCPNInfoList(void) const

{
    if (!CheckCPNInfoList())
        return 0;
    return Mutable->CPNInfoList;
}

// =================================================================================== 

int
TParallelControl::GetCurrentNrOfHosts(void) const

{
    if (!CheckHostInfoList())
        return 0;
    return (int) Mutable->HostInfoList->GetNrOfElements();
}

// =================================================================================== 

int
TParallelControl::GetCurrentNrOfCPN(void) const

{    
    if (!CheckCPNInfoList())
        return 0;
    return (int) Mutable->CPNInfoList->GetNrOfElements();
}

// =================================================================================== 

const THostInfoList *
TParallelControl::GetHostInfoList(void) const

{
    if (!CheckHostInfoList())
        return 0;
    return Mutable->HostInfoList;
}

// =================================================================================== 

TString
TParallelControl::GetInstanceName(void) const

{
    char Line[64];
        
    sprintf(Line, "%s:%s", IsMaster()? "M":"S",
            (Mutable AND Mutable->LocalCPNInfo)? (char *) Mutable->LocalCPNInfo->Name() : "?");
    return Line;
}

// =================================================================================== 

TCPNID
TParallelControl::GetLocalCPNID(void) const

{
    if (!CheckCPNInfoList())
        return 0;
    return Mutable->LocalCPNInfo->ID;
}

// =================================================================================== 

const TCPNInfo *
TParallelControl::GetLocalCPNInfo(void) const

{
    if (!CheckCPNInfoList())
        return 0;
    return Mutable->LocalCPNInfo;
}

// =================================================================================== 

const THostInfo *
TParallelControl::GetLocalHostInfo(void) const

{
    if (!CheckHostInfoList())   
        return 0;
    return Mutable->LocalHostInfo;
}

// =================================================================================== 

int
TParallelControl::GetMaxNrOfCPN(void) const

{
    int           MaxNrOfCPN = 0;
    THostInfoEnum CHost;

    if (!CheckHostInfoList())
        return -1;
    CHost.WalksForwardThrough(Mutable->HostInfoList);
    while (CHost())
    {
        MaxNrOfCPN += CHost->MaxNrOfCPN;
        CHost++;
    }
    return MaxNrOfCPN;
}

// =================================================================================== 

TCPNID
TParallelControl::GetParentCPNID(void) const

{
    return 0;
}


// =================================================================================== 

TTopology *
TParallelControl::GetPhysicalTopology(void) const

{
    if (!DeterminePhysicalTopology())
        return 0;
    return Mutable->PhysicalTopology;
}

// ===================================================================================

bool
TParallelControl::HandleEventCPNInfo(TBaseMessage *aMessage, int &Handled)

{
    TCPNID    CPNID;
    TCPNInfo *CPNInfo;

    if (!aMessage->Decode(CPNID))
        return FALSE;
    if ((CPNInfo = GetCPNInfoByID(CPNID)) == 0)
    {
        CPNInfo = new TCPNInfo(aMessage);
        if (CPNInfo->Failed())
            return FALSE;
        if (!AddCPNInfo(CPNInfo))
            return FALSE;
    }
    else
        if (!CPNInfo->Decode(aMessage))
            return FALSE;
    Handled = TRUE;
    return TRUE;
}

// =================================================================================== 

bool
TParallelControl::HandleEventQuit(TBaseMessage *, int &Handled)

{
    debug(Print(DM_PARALLEL, "received EVENT_QUIT"));
    QuitFlag = TRUE;
    Handled = EXIT_LOOP;
    return TRUE;
}

// =================================================================================== 

bool
TParallelControl::HandleEventSyncAck(TBaseMessage *theMessage, int &Handled)

{
    if (SyncAckMissing == 0)
    {
        Error("received superflous EVENT_SYNC_ACK from %X",
                theMessage->GetSender());
        return FALSE;
    }
    Handled = (--SyncAckMissing)? TRUE :  EXIT_LOOP;
    debug(Print(DM_PARALLEL, "received EVENT_SYNC_ACK from %lX (missing:%d)",
                theMessage->GetSender(), SyncAckMissing));
    return TRUE;    
}

// =================================================================================== 

bool
TParallelControl::HandleEventSyncReq(TBaseMessage *theMessage, int &)

{
    debug(Print(DM_PARALLEL, "HandleEventSyncReq: returning EVENT_SYNC_ACK"));
    if (!Communication->SendEvent(EVENT_SYNC_ACK, theMessage->GetSender(),
                                  MTAG_PARALLEL_CONTROL, MTYPE_EVENT))
    {
        Error("cannot send EVENT_SYNC_ACK");
        return FALSE;
    }
    //    Handled = EXIT_LOOP;
    return TRUE;
    
}

// =================================================================================== 

bool
TParallelControl::HandleEventSyncReqExitLoop(TBaseMessage *theMessage, int &Handled)

{
    debug(Print(DM_PARALLEL, "HandleEventSyncReq: returning EVENT_SYNC_ACK"));
    if (!Communication->SendEvent(EVENT_SYNC_ACK, theMessage->GetSender(),
                                  MTAG_PARALLEL_CONTROL, MTYPE_EVENT))
    {
        Error("cannot send EVENT_SYNC_ACK");
        return FALSE;
    }
    Handled = EXIT_LOOP;
    return TRUE;
    
}

// =================================================================================== 

void
TParallelControl::_Init(void)

{
    TConfiguration *C = Configuration;

    Flags = 0;
    Enrolled = FALSE;
    QuitFlag = FALSE;
    EventHandler= new TEventHandler<TParallelControl> (this);
    EventHandler->Register(EVENT_CPNINFO, &TParallelControl::HandleEventCPNInfo);
    EventHandler->Register(EVENT_SYNC_REQ, &TParallelControl::HandleEventSyncReq);
    EventHandler->Register(EVENT_SYNC_REQ_EXIT_LOOP, &TParallelControl::HandleEventSyncReqExitLoop);
    EventHandler->Register(EVENT_SYNC_ACK, &TParallelControl::HandleEventSyncAck);
    EventHandler->Register(EVENT_QUIT, &TParallelControl::HandleEventQuit);
    SetEventHandler(EventHandler);
    Mutable = new TParallelMutable;
    Mutable->ParentCPNInfo = 0;
    Mutable->LocalCPNInfo = 0;
    Mutable->LocalHostInfo = 0;
    Mutable->CPNInfoList = 0;
    Mutable->HostInfoList = 0;
    Mutable->PhysicalTopology = 0;
#ifdef SLAVES_NEED_PRINT_PREFIX
    Flags |= PC_FLAG_SLAVES_NEED_PRINT_PREFIX;
#endif
    C->Set(TConfigEntry(PARALLEL_GROUP, PARALLEL_NAME_USEGRID, DEFAULT_USEGRID));
    C->Set(TConfigEntry(PARALLEL_GROUP, PARALLEL_NAME_GRIDROWS, DEFAULT_GRIDROWS));
    C->Set(TConfigEntry(PARALLEL_GROUP, PARALLEL_NAME_GRIDCOLS, DEFAULT_GRIDCOLS));
}

// =================================================================================== 

bool
TParallelControl::IsMaster(void) const

{
    return TRUE;
}

/* =================================================================================== */

bool
TParallelControl::Quit(void) const

{    
    if (IsMaster())
    {
        TBaseMessage *Message = 0;
        
        debug(Print(DM_PARALLEL, "Quit: broadcasting EVENT_QUIT"));
        if (!Communication->GetNewMessage(Message, DEST_ALL_BUT_MYSELF, 
                                          MTAG_PARALLEL_CONTROL, MTYPE_EVENT))
        {
            Error("Quit: cannot get new message");
            return FALSE;
        }
        Message->Event(EVENT_QUIT);
        return Communication->Commit(Message);
    }
    else
    {
        if (QuitFlag)
            return TRUE;
        debug(Print(DM_PARALLEL, "Quit: waiting for EVENT_QUIT"));
        return Communication->MessageLoop();
    }
}
    
// =================================================================================== 

bool
TParallelControl::SendChangesInTopology(void)

{
    TBaseMessage *Message = 0;
    TCPNInfoEnum  CInfo;

    if (!Communication->GetNewMessage(Message, DEST_ALL_BUT_MYSELF,
                                      MTAG_PARALLEL_CONTROL, MTYPE_EVENT))
    {
        Error("SendChangesInTopology: cannot get new message");
        return FALSE;
    }
    CInfo.WalksForwardThrough(Mutable->CPNInfoList);
    while (CInfo())
    {
        if (!(Message->Event(EVENT_CPNINFO) AND
              Message->Encode(CInfo->ID) AND
              CInfo->Encode(Message)))
            return FALSE;
        CInfo->Flags &= ~CPNINFO_FLAG_MODIFIED;
        CInfo++;
    }
    return Communication->Commit(Message);
}

// =================================================================================== 

bool
TParallelControl::Sync(void)

{
    if (IsMaster())
        return Synchronize(DEST_ALL_BUT_MYSELF, TRUE);
    else        
        return Communication->MessageLoop();
}

// =================================================================================== 

bool
TParallelControl::Synchronize(TDestination Destination, bool ExitLoop)

{
    if (Destination.GetNrOfRecipients() == 0)
        return TRUE;
    debug(Print(DM_PARALLEL, "Synchronize: broadcasting EVENT_SYNC_REQ to %ld CPN",
                Destination.GetNrOfRecipients()));
    if (!Communication->SendEvent(ExitLoop? EVENT_SYNC_REQ_EXIT_LOOP:EVENT_SYNC_REQ, 
                                  Destination, MTAG_PARALLEL_CONTROL, MTYPE_EVENT))
    {
        Error("Synchronize: cannot send EVENT_SYNC_REQ/ACK");
        return FALSE;
    }
    SyncAckMissing = Destination.GetNrOfRecipients();
    debug(Print(DM_PARALLEL, "Sync: waiting for EVENT_SYNC_ACK"));
    while (SyncAckMissing)
        if (!Communication->MessageLoop())
        {
            Error("Synchronize: while waiting for EVENT_SYNC_ACK");
            return FALSE;
        }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
