/*      file     : parbalance/TDynamicBalancer.h
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TDynamicBalancer.h,v 1.8 2005/03/25 11:06:49 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

#ifndef TDYNAMICBALANCER_H
#define TDYNAMICBALANCER_H

// ***INCLUDES************************************************************************ 

#include <TBaseObject.h>
#include <TTimeStep.h>
#include <TMessageRecipient.h>
#include <TObjectHandler.h>
#include <TStatHandler.h>
#include <TTransferLoadInfo.h>

// ***FORWARDS************************************************************************

class TBaseBalancer;
class TExtendedLoadStat;
class TLoadStat;
class TMigrationContext;
class TRandom;
class TStatistics;
class TStream;
class TTimer;
class TTopology;

#include <TNode.forward.h>
#include <TEdge.forward.h>
#include <TNodeHeapEntry.forward.h>

// ***CLASSES************************************************************************* 

class TDynamicBalancer : public TBaseObject, public TMessageRecipient, 
                         public TObjectHandler, public TStatHandler

{
public:

    TDynamicBalancer(TNodeList *theNodes, TEdgeList *theEdges, 
                     TEdgeList *theBoundaryEdges, 
                     TTimer *theNodeTimer, TTimer *theEdgeTimer,
                     TTimer *theAppMiscTimer);
    virtual ~TDynamicBalancer(void);

    bool
    Activate(void);

    bool
    ExecuteSubTimeStep(TTimeStep &TimeStep);

    TLoadStat *
    GetLatestGlobalLoadStat(void);

    bool
    Run(void);

    bool
    StartNeighborCommStat(void);

private:

    bool
    ActivateStatistics(void);

    bool
    AddBoundaryNode(TMigrationContext &C,  TNode *InitialNode);

    bool
    AddNodeToContext(TMigrationContext &C, TNode *Node);

    bool
    CancelSend(TTransferLoadInfo *Info, const TTimeStep *TimeStep = 0);

    bool
    CheckLocalStats(void);

    bool
    CheckPendingTransfers(TTimeStep &TimeStep);

    bool
    CheckNewTransfers(TTimeStep &TimeStep);

    virtual bool
    Collect(TSamplingDescription *Description);

    bool
    CollectNeighborCommInfos(TStatObjectList *StatList);

    bool
    ComputeLocalLoadValues(void);

    bool
    DoubleCheckMigration(TTransferLoadInfo *Info);

    virtual TString
    GetClassName(void) const;

    virtual bool
    HandleCreateRequest(TParallelObject *&Object, const TClassInfo *ClassInfo, 
                        TParallelID ParallelID);

    virtual bool
    HandleEvent(TEventID EventID, TBaseMessage *aMessage, int &Handled);

    bool
    HandleEventAckSend(TBaseMessage *Message);

    bool
    HandleEventCancelSend(TBaseMessage *Message);

    bool
    HandleEventConfirmSend(TBaseMessage *Message);

    bool
    HandleEventDenySend(TBaseMessage *Message);

    bool
    HandleEventReady(TBaseMessage *Message, int &Handled);

    bool
    HandleEventRequestSend(TBaseMessage *Message);

    virtual bool
    HandleLoadRequest(TParallelObject *Object, bool &Loaded);

    bool
    HandleLoadStatResult(TAggregationInfo *AggregationInfo, TTimeStep &TimeStep);

    virtual bool
    HandleAggregationResult(TAggregationInfo *AggregationInfo, TTimeStep &TimeStep);

    bool
    HandleNeighborCommResult(TAggregationInfo *AggregationInfo, TTimeStep &TimeStep);

    bool
    HandleTimedEvent(TBaseMessage *Message, TTransferLoadInfo *&Info,
                     TTimeStep &TimeStep);

    bool
    Migrate(TTransferLoadInfo *Info);

    bool
    NewContext(TTransferLoadInfo *Info, TMigrationContext *&C);

    bool
    PrepareTransfer(TTimeStep &TimeStep, TTransferLoadInfo *Info);

    void
    RemoveBoundaryNode(TMigrationContext &C, TNode *Node);

    bool
    SelectNodes(TTransferLoadInfo *Info, TMigrationContext &C);

    bool
    SendTimedEvent(TDestination Destination, TEventID EventID,
                   const TTimeStep &TimeStep);

    void
    UpdateHeapEntry(TMigrationContext &C, TNodeHeapEntry *HeapEntry);

    void
    UpdateNodeHeap(TMigrationContext &C);

    bool
    Wait(TTransferLoadInfo *Info);

    bool
    WaitUntilReady(TTransferLoadInfo *Info);

    bool
    WriteLoadStat(TLoadStat *Stat);

    TTopology *Topology;
    TNodeList *Nodes;
    TEdgeList *Edges;
    TEdgeList *BoundaryEdges;

    TTimer    *NodeTimer;
    TTimer    *EdgeTimer;
    TTimer    *AppMiscTimer;
    TTimer    *ApplicationTimer;
    TTimer    *Timer;

    TRandom   *Random;
    int        SamplingInterval;
    int        HistoryLength;
    int        TransferDelay;
    int        TransferDelayVariation;

    TTransferLoadInfoList PendingTransfers;

    double     IdleTime;
    double     MiscTime;
    double     AppMiscTime;
    double     NodeTime;
    double     EdgeTime;
    double     ApplicationTime;
    double     NodeLoad;
    double     EdgeLoad;
    double     Perf;
    double     Load;
    double     ApplicationLoad;

    double TimeWeight;
    double AppLoadWeight;
    double LoadWeight;

    TStatistics *PerfHistory;
    int          PerfHistoryLength;
    TStatistics *TimeHistory;
    int          TimeHistoryLength;

    TBaseBalancer *Balancer;
    bool           UseLocal;
    bool           UsePLB;
    bool           Transfer;
    
    TLocation   MyCenterOfGravity;
    TRectangle  MyFrame;
    double      MyDiameter;

    TNodeHeap  *NodeHeap;
    int         PreviousNumberOfNodes;
    double      MinPenaltyChange;
    double      ExcessLoadWeight;
    double      LackOfLoadWeight;
    double      IncidenceWeight;
    double      DistanceWeight;

    TAggregationInfo     *NeighborCommAggregationInfo;
    TSamplingDescription *NeighborCommSamplingDescription;

    // global load statistics:

    TLoadStat             *LatestGlobalLoadStat;
    TSamplingDescription  *LoadSamplingDescription;
    TAggregationInfo      *LoadAggregationInfo;
    int                    LoadStatInterval;
    TStream               *LoadStatOutput;
    bool                   LoadStatDump;
};

// ***EXTERNALS*********************************************************************** 

extern TDynamicBalancer *DynamicBalancer;

// ===================================================================================
// Local Variables:
// mode:c++
// End:
#endif
