/*      file     : partools/TEdge.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TEdge.C,v 1.14 2005/05/01 14:57:58 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <stdio.h>

#include <TBaseStream.C>
#include <TObjectManager.h>

#include "TNode.h"

#include "TTransferInfo.h"
#include "TEdge.h"

// ***GLOBAL VARIABLES****************************************************************  

TClassInfo *TEdge::ClassInfo = 0;

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include "TEdge.i.C"
#endif

// =================================================================================== 
#define ParentClass TContainerObject
// =================================================================================== 

TEdge::TEdge(TClassInfo *theClassInfo, TParallelID theParallelID,
             TInfoID theDefaultAdjacenceInfoID) :

ParentClass(theClassInfo, theParallelID)

{
    DefaultAdjacenceInfoID = theDefaultAdjacenceInfoID;
}

// =================================================================================== 

bool
TEdge::Decode(TBaseStream *Stream, TTransferInfo *TransferInfo)

{
    if (TransferInfo == 0)
    {
        Error("no TransferInfo available");
        return FALSE;
    }
    if (TransferInfo->TransferMode == TM_ConstData)
        if (!TransferInfo->Decode(Stream))
            return FALSE;
    return ParentClass::Decode(Stream, TransferInfo);
}

// =================================================================================== 

bool
TEdge::DisposeNonLocalPortion(TCPNID CPNID)

{
    TDistributedRange  Range;
    int                CInfo;
    TParallelObject   *Info;

    GetActiveRange(CPNID, &Range);
    Range.Invert();
    for (CInfo = 0; CInfo < GetClassInfo()->GetNrOfInfoClasses(); CInfo++)
        if ((Info = GetInfo(CInfo)))
            if (!Info->DisposeRange(&Range))
            {
                Error("DisposeNonLocalPortion: cannot dispose range for %s",
                      (char *) Info->GetObjectName());
                return FALSE;
            }
    return TRUE;
}

// =================================================================================== 

bool
TEdge::Encode(TBaseStream *Stream, TTransferInfo *TransferInfo) const

{
    if (TransferInfo == 0)
    {
        Error("no TransferInfo available");
        return FALSE;
    }
    if (TransferInfo->TransferMode == TM_ConstData)
    {
        if (TransferInfo->Destination == GetFromNodeCPNID())
            TransferInfo->TransferRange.Add(0.0, 0.5);
        if (TransferInfo->Destination == GetToNodeCPNID())
            TransferInfo->TransferRange.Add(0.5, 1.0);
        if (!TransferInfo->Encode(Stream))
            return FALSE;
    }
    return ParentClass::Encode(Stream, TransferInfo);
}

// =================================================================================== 

double
TEdge::GetActiveFraction(TCPNID CPNID) const

{
    TDistributedRange Range;

    GetActiveRange(CPNID, &Range);
    return Range.GetFraction();
}

// =================================================================================== 

void
TEdge::GetActiveRange(TCPNID CPNID, TDistributedRange *Range) const

{
    Range->MarkInactive();
    if (GetFromNodeCPNID() == CPNID)
        Range->Add(0.0, 0.5);
    if (GetToNodeCPNID() == CPNID)
        Range->Add(0.5, 1.0);
}

// =================================================================================== 

TString
TEdge::GetClassName(void) const

{
    return "TEdge";
}

// =================================================================================== 

TCPNID
TEdge::GetFromNodeCPNID(void) const

{
    TNode *Node;

    if ((Node = GetFromNode()) == 0)
    {
        Warning("GetFromNodeCPNID: cannot retrieve node");
        return INVALID;
    }
    return Node->GetCPNID();
}

// =================================================================================== 

TCoordinate
TEdge::GetLength(TInfoID AdjacenceInfoID, TInfoID LocationInfoID)

{
    TAdjacenceInfo *Adj;

    if ((Adj = (TAdjacenceInfo *) GetInfo(AdjacenceInfoID, TRUE, TRUE)) == 0)
    {
        Error("GetLength: cannot retrieve adjacence info");
        return (TCoordinate) INVALID;
    }
    return (TCoordinate) Adj->GetLength(LocationInfoID);
}

// =================================================================================== 

bool
TEdge::GetRequiredContainers(int &NrOfRefs, TContainerReference *Referencess) 

{
    TAdjacenceInfo *Adj;

    if ((Adj = (TAdjacenceInfo *) GetInfo(DefaultAdjacenceInfoID, TRUE, TRUE)) == 0)
    {
        Error("GetRequiredObjects: cannot retrieve adjacence info");
        return FALSE;
    }
    Referencess[0].ClassID = Referencess[1].ClassID = TNode::ClassInfo->GetID();
    Referencess[0].ParallelID = Adj->GetFromNodeID();
    Referencess[1].ParallelID = Adj->GetToNodeID();
    NrOfRefs = 2;
    return TRUE;
}

// =================================================================================== 

TCPNID
TEdge::GetToNodeCPNID(void) const

{
    TNode *Node;

    if ((Node = GetToNode()) == 0)
    {
        Warning("GetToNodeCPNID: cannot retrieve node");
        return INVALID;
    }
    return Node->GetCPNID();
}

/* =================================================================================== */

TEdge *
TEdge::LoadInstance(TParallelID EdgeID, TInfoMask LocalInfoMask, bool LoadNodes)

{
    TEdge *Edge;

    if ((Edge = GetPtr(EdgeID)) == 0)
        if ((Edge = (TEdge *) ObjectManager->CreateInstance(ClassInfo, EdgeID,
                                                            LocalInfoMask, FALSE)) == 0)
            return 0;
    if (LoadNodes)
    {
        TAdjacenceInfo *Adj;

        if ((Adj = (TAdjacenceInfo *) Edge->GetInfo(TAdjacenceInfo::ClassInfo->GetInfoID(), 
                                                    TRUE, TRUE)) == 0)
            return 0;
        if (!(TNode::LoadInstance(Adj->GetFromNodeID()) AND
              TNode::LoadInstance(Adj->GetToNodeID())))
            return 0;
    }
    return Edge;
}

// =================================================================================== 
#undef ParentClass


