/*      file     : partools/TObjectManager.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TObjectManager.C,v 1.20 2005/05/01 14:57:58 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <TIDContainerEntry.h>
#include <TConfiguration.h>
#include <TCommunication.h>
#include <TCPNInfo.h>
#include <TBaseStream.h>

#include "ParToolsConfig.h"

#include "TExtendedCPNInfo.h"
#include "TCommunicationLink.h"
#include "TTransferInfo.h"
#include "TClassInfo.h"
#include "TBoundary.h"
/*
#include "TNode.h"
#include "TEdge.h"
#include "TIncidenceInfo.h"
*/
#include "TNetwork.h"
#include "TObjectManager.h"

// ***DEFINES************************************************************************* 

#define OBJ_MGR_DATALOCAL  0L

// ***GLOBAL VARIABLES****************************************************************  

TObjectManager *ObjectManager = 0;

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include "TObjectManager.i.C"
#endif

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TObjectManager::TObjectManager(void)

{
    if (ObjectManager != 0)
    {
        Fail("another instance already active");
    }
    ObjectManager = this;
#ifdef FAST_HEAP
    TParallelObject::InitHeapManager();
#endif
    if (Communication)
    {
        Communication->RegisterRecipient(MTAG_OBJ_MGR | MTAG_MIGRATION | 
                                         MTAG_BOUNDARIES, this);
        Communication->DeactivateTags(MTAG_BOUNDARIES);
    }
    Configuration->Set(TConfigEntry(OBJ_MGR_GROUP, OBJ_MGR_NAME_DATALOCAL,
                                    OBJ_MGR_DATALOCAL));
    ClassInfoArray = 0;
    DataReceived = FALSE;
    IncomingBoundaries[0] = &_IncomingBoundaries[0];
    IncomingBoundaries[1] = &_IncomingBoundaries[1];
    MissingBoundarySets[0] = 0;
    MissingBoundarySets[1] = 0;
    GlobalMigrationMonitor = INVALID;
    PreviousCPNIDOfLatestContainer = INVALID;
    PreviousObjectExisted = FALSE;
    ObjectManager->RegisterContainerClass(0, TExtendedCPNInfo::ClassInfo, 
                                          CLASS_EXTENDED_CPNINFO, this);
    ObjectManager->RegisterContainerClass(0, TCommunicationLink::ClassInfo, 
                                          CLASS_COMMUNICATION_LINK, this);
    debug(Print(DM_LIFESPAN, "CONSTRUCTOR"));
}

// =================================================================================== 

TObjectManager::~TObjectManager(void)

{
    debug(Print(DM_LIFESPAN, "DESTRUCTOR"));
    DeleteObjectArray(ClassInfoArray);
#ifdef FAST_HEAP
    TParallelObject::DumpHeapManagerStat();
#endif
    ObjectManager = 0;
}

// =================================================================================== 

bool
TObjectManager::Activate(void)

{
    debug(Print(DM_LIFESPAN, "Activate"));
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::ActivateClasses(void)
    
{
    TEnumerator< TClassInfo > CInfo(ClassInfos);
    TClassID                  CClassID;

    debug(Print(DM_CLASS, "activate classes:"));
    CClassID = 0;
    ClassInfoArray = new TClassInfo * [ClassInfos.GetNrOfElements()];
    CInfo.Enumerate();
    while(CInfo())
    {
        if (CInfo->ContainerClassname != 0)
        {
            TSelector< TClassInfo , TClassInfo > SearchInfo(ClassInfos);

            SearchInfo.Select(TClassInfo(CInfo->ContainerClassname));

            if (!SearchInfo())
            {
                Error("ActivateClasses: cannot find container class for %s",
                      (char *) CInfo->GetInstanceName());
                return FALSE;
            }
            CInfo->ContainerClass = SearchInfo;
        }
        CInfo++;
    }
    CInfo.Enumerate();
    while(CInfo())
    {
        CInfo->ClassID = CClassID;
        ClassInfoArray[CClassID] = CInfo;
        if (CInfo->ContainerClass)
            CInfo->ContainerClass->NrOfInfoClasses++;
        CClassID++;
        CInfo++;
    }
    CInfo.Enumerate();
    while(CInfo())
    {
        if (CInfo->IsContainer())
        {
            CInfo->InfoClasses = new TClassInfo * [CInfo->NrOfInfoClasses];
            CInfo->NrOfInfoClasses = 0;
        }
        CInfo++;
    }
    CInfo.Enumerate();
    while(CInfo())
    {
        if (CInfo->ContainerClass)
        {
            if (!CInfo->ContainerClass->IsContainer())
            {
                Error("%s is not a container class", 
                      (char *) CInfo->ContainerClass->GetInstanceName());
                return FALSE;
            }
            CInfo->ContainerClass->InfoClasses[CInfo->ContainerClass->NrOfInfoClasses] = CInfo;
            CInfo->InfoID = CInfo->ContainerClass->NrOfInfoClasses++;
        }
        debug(Print(DM_CLASS, "class '%s'", (char *) CInfo->GetInstanceName()));
        CInfo++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::Action(TClassInfo *ClassInfo, TParallelObject *Object, 
                       TObjectAction Action)

{
    TObjectHandlerEnum CHandler;
    
    CHandler.WalksForwardThrough(ClassInfo->GetActionHandlers());
    while (CHandler())
    {
        if (Object)
        {
            if (!CHandler->HandleAction(Object, Action))
            {
                Error("Action: action %d for object %s", (int) Action, 
                      (char *) Object->GetInstanceName());
                return FALSE;
            }
        }
        else
            if (!CHandler->HandleAnonymousAction(ClassInfo, Action))
            {
                Error("Action: anonymous action %d", (int) Action);
                return FALSE;
            }
        CHandler++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::AddBoundaryContainer(TContainerObject *Object, TCPNID RemoteCPNID)

{
    TExtendedCPNInfo *Info;
    
    if ((Info = Network->GetExtendedCPNInfo(RemoteCPNID, TRUE)) == 0)
    {
        Error("AddBoundaryContainer: cannot add ExtendedCPNInfo "
              "for boundary %s", (char *) Object->GetObjectName());
        return FALSE;
    }
    Info->BoundaryContainers.Append(Object);
    if (!Action(Object->GetClassInfo(), Object, ActionAddBoundary))
        return FALSE;
    if (!Neighbors.IsInList(Info->CPNInfo))
    {
        IfDebug(DM_TOPOLOGY, Print("add nghbr %lX", Info->CPNInfo->ID));
        Neighbors.Append(Info->CPNInfo);
    }
    IfDebug(DM_TOPOLOGY, Print("add bnd cntnr %s to %lX", 
                               (char *) Object->GetObjectName(), Info->CPNInfo->ID));
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::AddContainer(TContainerObject *Object)

{
    if (!LocalContainerObjects.IsInList(Object))
    {
        IfDebug(DM_CONTAINER, Print("add container %s", (char *) Object->GetObjectName());
                Object->AddInstance());
        if (!Action((TClassInfo *) Object->GetClassInfo(), Object, ActionAdd))
            return FALSE;
        LocalContainerObjects.Append(Object);
    }
    else
    {
        IfDebug(DM_CONTAINER, Print("merge container %s", (char *) Object->GetObjectName()));
        if (!Action((TClassInfo *) Object->GetClassInfo(), Object, ActionMerge))
            return FALSE;       
    }
    NewContainerObjects.Append(Object);
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::AddContainersInIndex(TClassInfo *ClassInfo, TParallelID *Index, 
                                     long Count, TContainerObjectList *List,
                                     TInfoMask LoadableMask, TInfoMask RequiredMask,
                                     TInfoMask LocalMask)

{
    TContainerObject *Object;
    int               i;

    for (i = 0; i < Count; i++)
    {
        if ((Object = (TContainerObject *) CreateInstance(ClassInfo, Index[i],
                                                          LocalMask, TRUE)) == 0)
        {
            Error("AddContainersInIndex: cannot create object (index[%ld]=%ld)",
                  i, Index[i]);
            return FALSE;
        }
        Object->SetLoadableMask(LoadableMask);
        Object->SetRequiredMask(RequiredMask);
        List->Append(Object);
        if (LocalMask == 0)
            if (!AddContainer(Object))
            {
                Error("AddContainersInIndex: cannot add %s", 
                      (char *) Object->GetObjectName());
                return FALSE;
            }
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::AddNewBoundaries(TContainerObject *Object, TCPNID RemoteCPNID)

{
    TContainerReference  Refs[MAX_CONTAINER_DEPS];
    int  NrOfRefs;    
    int  CRef;
    bool IsPartiallyLocal = FALSE;
    bool IsLocal = TRUE;

    IfDebug(DM_BOUNDARY, Print("CheckNewBoundaries: %s to CPN %lX",
                               (char *) Object->GetObjectName(), RemoteCPNID));

    if (!Object->GetRequiredContainers(NrOfRefs, Refs + 0))
    {
        Error("CheckNewBoundaries: cannot retrieve refs for %s",
              (char *) Object->GetObjectName());
        return FALSE;
    }
    for (CRef = 0; CRef < NrOfRefs; CRef++)
    {
        TContainerObject *RequiredContainer;
        
        if ((RequiredContainer = GetContainer(Refs[CRef], RemoteCPNID)) == 0)
        {
            Error("CheckNewBoundaries: cannot retrieve required container %d:%ld",
                  Refs[CRef].ClassID, Refs[CRef].ParallelID);
            return FALSE;
        }
        if (RequiredContainer->IsLocal(Parallel->GetLocalCPNID()))
            IsPartiallyLocal = TRUE;
        else
            IsLocal = FALSE;
    }
    if (!IsLocal)
    {
        if (IsPartiallyLocal)
        {
            for (CRef = 0; CRef < NrOfRefs; CRef++)
            {
                TContainerObject *RequiredContainer;
                
                if ((RequiredContainer = GetContainer(Refs[CRef])) == 0)
                {
                    Error("CheckNewBoundaries: cannot retrieve required container");
                    return FALSE;
                }
                if (!RequiredContainer->IsLocal(Parallel->GetLocalCPNID()))
                {
                    AddRemoteContainer(RequiredContainer);
                    if (!AddBoundaryContainer(Object, RequiredContainer->GetCPNID()))
                    {
                        Error("CheckNewBoundaries: cannot add boundary");
                        return FALSE;
                    }
                }
            }
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::AddRemoteContainer(TContainerObject *Object)

{
    if (!RemoteContainers.IsInList(Object))
    {
        RemoteContainers.Append(Object);
        Object->AddInstance();
        IfDebug(DM_MIGRATION, Print("add remote container %s", 
                                    (char *) Object->GetObjectName()));
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::AssignInfo(TContainerObject *Object, const TParallelObject *Info)

{
    TClassInfo       *InfoClassInfo;
    TParallelObject  *PreviousInfo;

    //    Print("AssignInfo '%s' to container '%s'", (char *) Info->GetObjectName(),
    //  (char *) Object->GetObjectName());
    if (Object->GetParallelID() != Info->GetParallelID())
    {
        Error("AssignInfo: ParallelIDs do not match");
        return FALSE;
    }
    InfoClassInfo = (TClassInfo *) Info->GetClassInfo();
    if ((PreviousInfo = Object->GetInfo(InfoClassInfo->GetInfoID())))
    {
        Warning("AssignInfo: destroying previous info '%s'",
                (char *) PreviousInfo->GetInstanceName());
        if (!DestroyObject(PreviousInfo))
            return FALSE;
    }
    Object->SetInfo(InfoClassInfo->GetInfoID(), Info);
    return TRUE;
}

// =================================================================================== 

TContainerObject *
TObjectManager::CreateInstance(const TClassInfo *ClassInfo, TParallelID ParallelID,
                               TInfoMask LocalMask, bool CountInstance)

{
    TContainerObject *Object;

    if ((Object = (TContainerObject *) ClassInfo->GetObjectPtr(ParallelID)) == 0)
    {
        if ((Object = (TContainerObject *) CreateObject(ClassInfo, ParallelID)) == 0)
        {
            Error("CreateInstance: cannot create");
            return FALSE;
        }
        Object->AddInstance(LocalMask);
    }
    else
        if (CountInstance)
            Object->AddInstance(LocalMask);
    IfDebug(DM_CONTAINER, Print("create instance %s", (char *) Object->GetObjectName()));
    return Object;
}

// =================================================================================== 

TParallelObject *
TObjectManager::CreateObject(const TClassInfo *ClassInfo, TParallelID ParallelID)

{
    TParallelObject *Object = 0;

    if (!ClassInfo->GetObjectHandler()->
        HandleCreateRequest(Object, ClassInfo, ParallelID))
    {
        Error("CreateObject: error while creating '%s' for ParallelID %ld", 
              (char *) ClassInfo->GetInstanceName(), ParallelID);
        return 0;
    }
    if (Object)
    {
        IfDebug(DM_OBJECT, Print("create %s", (char *) ClassInfo->GetInstanceName()));
    }
    else
        IfDebug(DM_OBJECT, Print("cannot create '%s' for ParallelID %ld",
                                 (char *) ClassInfo->GetInstanceName(), ParallelID));
    return Object;
}

// =================================================================================== 

bool
TObjectManager::DestroyInstance(TContainerObject *Object)

{
    IfDebug(DM_CONTAINER, Print("DestroyInstance %s", (char *) Object->GetObjectName()));
    Object->RemoveInstance();
    if (Object->GetNrOfInstances() == 0)
        return DestroyObject(Object);
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::DestroyObject(TParallelObject *Object)

{
    IfDebug(DM_OBJECT, Print("destroy %s", (char *) Object->GetObjectName()));
    if (!Action((TClassInfo *) Object->GetClassInfo(), Object, ActionDestroy))
        return FALSE;
    delete Object;
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::DetermineDependantContainers(TContainerObjectList *Containers,
                                             TContainerObjectList *Dependants)

{
    TContainerObjectEnum CObject;
    TContainerReference  Refs[MAX_CONTAINER_DEPS];

    CObject.WalksForwardThrough(Containers);
    while (CObject())
    {
        int NrOfRefs;
        int CRef;

        if (!CObject->GetDependantContainers(NrOfRefs, Refs + 0))
        {
            Error("DetermineDependantContainers: cannot retrieve refs for %s",
                  (char *) CObject->GetObjectName());
            return FALSE;
        }
        Print(DM_MIGRATION, "%s has %d dependants", 
              (char *) CObject->GetObjectName(), NrOfRefs);         
        for (CRef = 0; CRef < NrOfRefs; CRef++)
        {
            TContainerObject *Dependant;

            if ((Dependant = GetContainer(Refs[CRef])) == 0)
            {
                Error("DetermineDepedantContainers: cannot retrieve dependant object");
                return FALSE;
            }
            Print(DM_MIGRATION, "add dependant %s", (char *) Dependant->GetObjectName());         
            Dependants->Append(Dependant);
        }
        CObject++;
    }    
    IfDebug(DM_MIGRATION, Print("%ld dependant(s) found for %ld container(s)",
                                Dependants->GetNrOfElements(), Containers->GetNrOfElements()));
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::FillContainer(TContainerObject *ContainerObject,
                              TInfoMask RequiredMask)

{
    TInfoID          InfoID;
    TInfoMask        Bit;
    TParallelObject *Info;
    TClassInfo      *ContainerClassInfo;
    TInfoMask        LoadableMask;

    if (RequiredMask == 0)
    {
        LoadableMask = ContainerObject->GetLoadableMask();
        RequiredMask = ContainerObject->GetRequiredMask();
    }
    else
        LoadableMask = RequiredMask;
    ContainerClassInfo = GetClassInfo(ContainerObject->GetClassID());
    for ActiveBitsInMask(InfoID, Bit, LoadableMask)
    {
        bool        Loaded = FALSE;
        TClassInfo *InfoClassInfo;
        
        InfoClassInfo = ContainerClassInfo->GetInfoClassInfo(InfoID);
        if ((Info = ContainerObject->GetInfo(InfoID)) == 0)
        {
            
            Info = CreateObject(InfoClassInfo, 
                                ContainerObject->GetParallelID());
            if (Info == 0 AND (RequiredMask & Bit))
            {
                Error("FillContainer: cannot create required '%s'",
                      (char *) InfoClassInfo->GetInstanceName());
                return FALSE;
            }
            if (Info)
            {
                ContainerObject->SetInfo(InfoID, Info);
                if (!LoadObject(InfoClassInfo, Info, Loaded))
                {
                    Error("FillContainer: error while loading '%s' for ParallelID %lld", 
                          (char *) InfoClassInfo->GetInstanceName(), 
                          ContainerObject->GetParallelID());
                    return FALSE;
                }
                if (!Loaded)
                {
                    if (RequiredMask & Bit)
                    {
                        Error("FillContainer: cannot load required '%s' for ParallelID %lld", 
                              (char *) InfoClassInfo->GetInstanceName(), 
                              ContainerObject->GetParallelID());
                        return FALSE;
                    }
                    Warning("FillContainer: cannot load '%s' for ParallelID %lld",
                            (char *) InfoClassInfo->GetInstanceName(), 
                            ContainerObject->GetParallelID());
                }
            }
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::FillContainers(TContainerObjectList *ContainerObjectList,
                               TInfoMask InfoMask)

{
    TContainerObjectEnum CObject;

    CObject.WalksForwardThrough(ContainerObjectList);
    while (CObject())
    {
        if (!FillContainer(CObject, InfoMask))
            return FALSE;
        CObject++;
    }
    return TRUE;
}

// =================================================================================== 

TClassInfo *
TObjectManager::GetClassInfoByName(const char *Name)

{
    TSelector< TClassInfo , TClassInfo > SearchInfo(ClassInfos);

    SearchInfo.Select(TClassInfo(Name));    
    if (SearchInfo())
        return SearchInfo;
    else
        return 0;
}

// =================================================================================== 

TString
TObjectManager::GetClassName(void) const

{
    return "TObjectManager";
}

// =================================================================================== 

TContainerObject *
TObjectManager::GetContainer(TContainerReference &Ref, TCPNID CPNID)

{
    TClassInfo       *ClassInfo;
    TContainerObject *Object;

    if ((ClassInfo = GetClassInfo(Ref.ClassID)) == 0)
    {
        Warning("GetContainer: invalid ClassID %d", Ref.ClassID);
        return 0;
    }
    if ((Object = (TContainerObject *) ClassInfo->GetObjectPtr(Ref.ParallelID)) == 0)   
        if (CPNID != INVALID)
        {
            if ((Object = (TContainerObject *) CreateObject(ClassInfo, Ref.ParallelID)) == 0)
            {
                Error("GetContainer: cannot create %s for ParallelID %ld",
                      (char *) ClassInfo->GetInstanceName(),  Ref.ParallelID);
                return 0;
            }
            Object->SetCPNID(CPNID);
        }
    return Object;
}

// =================================================================================== 

TString
TObjectManager::GetInstanceName(void) const

{
    char Line[256];

    sprintf(Line, "C:%ld", LocalContainerObjects.GetNrOfElements());
    return Line;
}

// =================================================================================== 

bool
TObjectManager::HandleEvent(TEventID EventID, TBaseMessage *Message, int &Handled)

{
    switch(EventID)
    {
    case MEVENT_OBJ_MGR_BOUNDARIES:
        Handled = TRUE;
        return HandleEventBoundaries(Message, Handled);

    case MEVENT_OBJ_MGR_TRANSFER_CONTAINER:
        Handled = TRUE;
        return ReceiveContainerObject(Message);

    case MEVENT_OBJ_MGR_MIGRATION_INFO:
        Handled = TRUE;
        return ReceiveMigrationInfo(Message, TRUE);

    case MEVENT_OBJ_MGR_MONITOR_MIGRATION_INFO:
        Handled = TRUE;
        return ReceiveMigrationInfo(Message, FALSE);

    case MEVENT_OBJ_MGR_TRANSFER_COMPLETE:
        Handled = EXIT_LOOP;
        return InitializeDataAfterReception(Message->GetSender());

    default:
        Error("invalid event ID %ld", EventID);
        return FALSE;
    }
#ifndef COMPLETE_SWITCH_FLOW
    return TRUE;
#endif
}

// =================================================================================== 

bool
TObjectManager::HandleEventBoundaries(TBaseMessage *Message, int &Handled)

{
    TTransferInfo     TransferInfo;
    TTimeStep         TimeStep;
    TBoundaryList     Boundaries;
    TBoundaryEnum     CBoundary;
    TExtendedCPNInfo *Info;
    long              DeltaTime;

    if (!(TimeStep.Decode(Message) AND
          TransferInfo.Decode(Message) AND
          ReceiveObjectList((TParallelObjectList *) &Boundaries, 
                            Message, FALSE, &TransferInfo)))      
    {
        Error("HandleEventBoundaries: cannot receive boundaries");
        return FALSE;
    }
    CBoundary.WalksForwardThrough(Boundaries);
    while (CBoundary())
    {
        CBoundary->SetSender(Message->GetSender());
        CBoundary++;
    }
    if ((Info = Network->GetExtendedCPNInfo(Message->GetSender(), FALSE)) == 0)
    {
        Error("HandleEventBoundaries: cannot retrieve ExtendedCPNInfo");
        return FALSE;
    }
    DeltaTime = TimeStep - BoundaryTimeStep;
    if (DeltaTime < 0 OR DeltaTime > MAX_TIME_LAG)
    {
        Error("HandleEventBoundaries: received time-step %s while waiting for %s",
              (char *) TimeStep.GetInstanceName(), 
              (char *) BoundaryTimeStep.GetInstanceName());
        return FALSE;
    }
    IncomingBoundaries[DeltaTime]->Append(Boundaries);
    MissingBoundarySets[DeltaTime]--;
    if (DeltaTime == 0)
    {
        /*
        if (MissingBoundarySets[DeltaTime] < 0)
        {
            Error("received superfluous boundaries from %lX", Info->CPNInfo->ID);
            return FALSE;
        }
        */
        if (MissingBoundarySets[DeltaTime] == 0)
            Handled = EXIT_LOOP;
        IfDebug(DM_BOUNDARY, Print("received %ld boundaries for time-step %s from %lX "
                                   " (sets missing:%ld)", Boundaries.GetNrOfElements(), 
                                   (char *) TimeStep.GetInstanceName(), 
                                   Info->CPNInfo->ID, MissingBoundarySets[DeltaTime]));
    }
    else
    {
        IfDebug(DM_BOUNDARY, Print("received %ld early boundaries for time-step %s "
                                   "from %lX (sets stored:%ld)",Boundaries.GetNrOfElements(), 
                                   (char *) TimeStep.GetInstanceName(), Info->CPNInfo->ID, 
                                   -MissingBoundarySets[DeltaTime]));
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::InitializeDataAfterReception(TCPNID FromCPNID)

{
    TContainerObjectEnum  CObject;

    IfDebug(DM_MIGRATION, Print("initialize data after reception from %lX", FromCPNID));
    if (!RemoveOldBoundaries(FromCPNID))
    {
        Error("InitDataAfterReception: while removing old boundaries");
        return FALSE;
    }
    CObject.WalksForwardThrough(NewContainerObjects);
    while (CObject())
    {
        if (!AddNewBoundaries(CObject, FromCPNID))
        {
            Error("InitDataAfterReception: cannot add new boundaries for '%s'",
                  (char *) CObject->GetObjectName());
            return FALSE;
        }
        CObject++;
    }
    if (NewContainerObjects.GetNrOfElements() > 0)
    {
        IfDebug(DM_CONTROL, Print("initialized %ld container(s)",
                                  NewContainerObjects.GetNrOfElements()));
        NewContainerObjects.DeleteAll();
    }
    DataReceived = TRUE;
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::LoadDataLocally(void) const

{
    return Configuration->GetLong(OBJ_MGR_GROUP, OBJ_MGR_NAME_DATALOCAL);
}

// =================================================================================== 

bool
TObjectManager::LoadObject(TClassInfo *ClassInfo,
                           TParallelObject *Object, bool &Loaded)

{
    if (!ClassInfo->GetObjectHandler()->HandleLoadRequest(Object, Loaded))
    {
        Error("LoadObject: error while loading info '%s'", 
              (char *) ClassInfo->GetInstanceName());
        return FALSE;
    }
    if (Loaded)
    {
        IfDebug(DM_OBJECT, Print("loaded info '%s'", (char *) Object->GetObjectName()));
        if (!Action(ClassInfo, Object, ActionLoaded))
            return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::Migrate(bool)

{
    TEnumerator < TExtendedCPNInfo > CCPN(TExtendedCPNInfo::ClassInfo->GetTable());

    CCPN.Enumerate();
    while (CCPN())
    {
        TBaseMessage *Message = 0;
        
        if (CCPN->CPNInfo->ID != Parallel->GetLocalCPNID())
        {
            if (CCPN->ContainersToBeMigrated.GetNrOfElements() > 0)
            {
                TContainerObjectEnum CObject;

                if (!Communication->GetNewMessage(Message, CCPN->CPNInfo->ID, 
                                                  MTAG_MIGRATION, 
                                                  MTYPE_OBJ_MGR_MIGRATION))
                {
                    Error("MigrateContainerObjects: cannot get new message");
                    return FALSE;
                }
                IfDebug(DM_CONTROL, Print("migrate %ld container(s) to CPN %lX", 
                                          CCPN->ContainersToBeMigrated.GetNrOfElements(), 
                                          (char *) CCPN->CPNInfo->ID));
                CObject.WalksForwardThrough(CCPN->ContainersToBeMigrated);
                while (CObject())
                {
                    if (!MigrateContainerObject(Message, CCPN->CPNInfo->ID, CObject))
                        return FALSE;
                    CObject++;
                }
                if (!Communication->Commit(Message))
                {
                    Error("MigrateContainerObjects: cannot commit message");
                    return FALSE;
                }
            }
            if (!RemoveOldBoundaries(CCPN->CPNInfo->ID))
            {
                Error("InitDataAfterReception: while removing old boundaries");
                return FALSE;
            }
        }    
        TransferContainers.Append(CCPN->ContainersToBeMigrated);
        RequiredContainers.Append(CCPN->RequiredToBeMigrated);
        DependantContainers.Append(CCPN->DependantsToBeMigrated);
        CCPN->ContainersToBeMigrated.DeleteAll();
        CCPN->DependantsToBeMigrated.DeleteAll();
        CCPN->RequiredToBeMigrated.DeleteAll();
        CCPN++;
    }
    if (!SendMigrationInfos())
    {
        Error("Migrate: cannot send migrations infos to global monitor");
        return FALSE;
    }
    Print(DM_MIGRATION, "Migrate: synchronize with %ld transfer destinations",
          TransferDestinations.GetNrOfElements());
    if (!Communication->SendEvent(MEVENT_OBJ_MGR_TRANSFER_COMPLETE, &TransferDestinations,
                                  MTAG_MIGRATION, MTYPE_OBJ_MGR))
    {
        Error("Migrate: while sending MEVENT_OBJ_MGR_TRANSFER_COMPLETE");
        return FALSE;
    }
    if (!RemoveNonLocalInstances())
    {
        Error("Migrate: while removing non-local container objects");
        return FALSE;
    }
    TransferDestinations.DeleteAll();
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::MigrateContainerObject(TBaseStream *Stream, TCPNID Dest, 
                                       TContainerObject *Object)

{
    TClassInfo    *ClassInfo;
    TTransferInfo  TransferInfo;
    TInfoID        CInfo;
    TInfoMask      Mask;

    if (!Stream->Event(MEVENT_OBJ_MGR_TRANSFER_CONTAINER))
    {
        Error("TransferContainer: cannot send MEVENT_OBJ_MGR_TRANSFER_CONTAINER");
        return FALSE;
    }
    TransferInfo.TransferMode = TM_ConstData;
    TransferInfo.Destination = Dest;
    ClassInfo = Object->GetClassInfo();
    if (!SendObject(Object, Stream, &TransferInfo))
    {
        Error("TransferContainer: cannot send '%s'", (char *) ClassInfo->GetInstanceName());
        return FALSE;
    }   
    Mask = 1;
    for (CInfo = 0; CInfo < ClassInfo->GetNrOfInfoClasses(); CInfo++)
    {
        TParallelObject *Info = 0;
        
        if ((Object->GetLoadableMask() & Mask) AND (Info = Object->GetInfo(CInfo)))
        {
            TransferInfo.TransferMode = TM_ConstData;
            if (!Stream->Encode(CInfo))
                return FALSE;
            if (!LoadDataLocally())
                if (!Info->Encode(Stream, &TransferInfo))
                {
                    Error("TransferContainer: cannot encode const data of '%s'",
                          (char *) ClassInfo->GetInfoClassInfo(CInfo)->GetInstanceName());
                    return FALSE;           
                }
            TransferInfo.TransferMode = TM_VarData;
            if (!Info->Encode(Stream, &TransferInfo))
            {
                Error("TransferContainer: cannot encode var data of '%s'",
                      (char *) ClassInfo->GetInfoClassInfo(CInfo)->GetInstanceName());
                return FALSE;       
            }       
        }
        Mask = Mask << 1;
    }
    CInfo = INVALID;
    if (!Stream->Encode(CInfo))
        return FALSE;
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::MigrateContainerObjects(const TCPNInfo *Destination, 
                                        TContainerObjectList *List)

{
    TContainerObjectEnum  CObject;
    TExtendedCPNInfo     *Info;
    
    if ((Info = Network->GetExtendedCPNInfo(Destination->ID, TRUE)) == 0)
    {
        Error("MigrateContainerObjects: cannot retrieve extended CPNInfo");
        return FALSE;
    }
    Info->ContainersToBeMigrated.Append(*List);
    Info->RequiredToBeMigrated.Append(*List);
    TransferDestinations.Append(Destination);
    if (!DetermineDependantContainers(&Info->RequiredToBeMigrated, 
                                      &Info->DependantsToBeMigrated))
    {
        Error("MigrateContainerObjects: cannot determine dependants");
        return FALSE;
    }
    Info->ContainersToBeMigrated.Append(Info->DependantsToBeMigrated);
    CObject.WalksForwardThrough(*List);
    while (CObject())
    {   
        CObject->SetCPNID(Destination->ID);
        CObject++;
    }
    Print(DM_CONTROL, "%ld containers selected to be migrated", 
          Info->ContainersToBeMigrated.GetNrOfElements());
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::ReceiveBoundaries(void)

{
    TBoundaryEnum  CBoundary;
    TBoundaryList *Temp;

    debug(Print(DM_BOUNDARY, "ReceiveBoundaries: time-step %s (missing sets: %ld)",
                (char *) BoundaryTimeStep.GetInstanceName(),
                MissingBoundarySets[0]));
    Communication->ActivateTags(MTAG_BOUNDARIES);
    while (MissingBoundarySets[0] > 0)
        if (!Communication->MessageLoop(TRUE))
        {
            Error("ReceiveBoundaries: while waiting in message loop");
            return FALSE;
        }
    Communication->DeactivateTags(MTAG_BOUNDARIES);
    debug(Print(DM_BOUNDARY, "ReceiveBoundaries: received all sets"));
    CBoundary.WalksForwardThrough(IncomingBoundaries[0]);
    while (CBoundary())
    {
        if (!CBoundary->Assign())
        {
            Error("ReceiveBoundaries: cannot assign %s", 
                  (char *) CBoundary->GetObjectName());
            return FALSE;
        }
        CBoundary++;
    }
    debug(Print(DM_BOUNDARY, "ReceiveBoundaries: assigned all %ld",
                IncomingBoundaries[0]->GetNrOfElements()));
    IncomingBoundaries[0]->DeleteAll(TRUE);
    /*
    Temp = IncomingBoundaries[0];
    IncomingBoundaries[1] = IncomingBoundaries[0];
    IncomingBoundaries[0] = Temp;
    */

    Temp = IncomingBoundaries[0];
    IncomingBoundaries[0] = IncomingBoundaries[1];
    IncomingBoundaries[1] = Temp;

    MissingBoundarySets[0] = MissingBoundarySets[1];
    MissingBoundarySets[1] = 0;
    BoundaryTimeStep++;
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::ReceiveContainerObject(TBaseStream *Stream)

{
    TClassInfo       *ClassInfo;
    TParallelID       ParallelID;
    TTransferInfo     TransferInfo;
    TContainerObject *Object;
    TInfoID           CInfo;

    TransferInfo.TransferMode = TM_ConstData;
    if ((Object = (TContainerObject *) ReceiveObject(Stream, TRUE, &TransferInfo)) == 0)
    {
        Error("ReceiveContainerObject: cannot receive");
        return FALSE;
    }
    IfDebug(DM_MIGRATION,Print("%s %s (PrevCPNID:%lX, was %sin CL)",
                               PreviousObjectExisted? "replaced":"received new",
                               (char *) Object->GetObjectName(), PreviousCPNIDOfLatestContainer,
                               LocalContainerObjects.IsInList(Object)? "":"not "));
    ParallelID = Object->GetParallelID();
    ClassInfo = Object->GetClassInfo();
    if (!Action(ClassInfo, Object, ActionDecoded))
        return FALSE;
    if (!(Stream->Decode(CInfo)))
    {
        Error("ReceiveContainerObject: cannot decode first InfoID");
        return FALSE;
    }
    while (CInfo != INVALID)
    {
        TClassInfo      *InfoClassInfo;
        TParallelObject *Info;
        bool             Loaded = FALSE;

        if (CInfo < 0 OR CInfo > ClassInfo->GetNrOfInfoClasses())
        {
            Error("ReceiveContainer: invalid info ID %d", CInfo);
            return FALSE;
        }
        InfoClassInfo = ClassInfo->GetInfoClassInfo(CInfo);
        if ((Info = Object->GetInfo(CInfo)) == 0)
        {
            
            if ((Info = CreateObject(InfoClassInfo, ParallelID)) == 0)
            {
                Error("ReceiveContainer: cannot create info '%s'",
                      (char *) InfoClassInfo->GetInstanceName());
                return FALSE;
            }
            Object->SetInfo(CInfo, Info);
        }
        if (LoadDataLocally())
        {
            if (!LoadObject(InfoClassInfo, Info, Loaded))
            {
                Error("ReceiveContainer: error while loading info '%s' for ParallelID %ld", 
                      (char *) InfoClassInfo->GetInstanceName(), ParallelID);
                return FALSE;
            }
            if (!Loaded)
            {
                Error("ReceiveContainer: cannot load info '%s'", 
                      (char *) InfoClassInfo->GetInstanceName());
                return FALSE;
            }
        }
        else
        {
            TransferInfo.TransferMode = TM_ConstData;
            if (!Info->Decode(Stream, &TransferInfo))
            {
                Error("ReceiveContainer: cannot decode const data of '%s'",
                      (char *) InfoClassInfo->GetInstanceName());
                return FALSE;
            }
            if (!Action(InfoClassInfo, Info, ActionDecoded))
                return FALSE;
        }
        TransferInfo.TransferMode = TM_VarData;
        if (!Info->Decode(Stream, &TransferInfo))
        {
            Error("ReceiveContainer: cannot decode var data of '%s'",
                  (char *) InfoClassInfo->GetInstanceName());
            return FALSE;
        }       
        if (!Stream->Decode(CInfo))
        {
            Error("ReceiveContainerObject: cannot decode subsequent InfoID");
            return FALSE;
        }
    }
    if (!AddContainer(Object))
    {
        Error("ReceiveContainerObject: cannot add %s", (char *) Object->GetObjectName());
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::ReceiveMigrationInfo(TBaseStream *Stream, bool CheckBoundaries)

{
    TTransferInfo        TransferInfo;
    TParallelObject     *Object;

    TransferInfo.TransferMode = TM_MigrationInfo;
    if ((Object = ReceiveObject(Stream, TRUE, &TransferInfo)) == 0)
    {
        Error("ReceiveMigrationInfo: cannot create object");
        return FALSE;
    }

    IfDebug(DM_MIGRATION, Print("received migration info for %s",
                                (char *) Object->GetObjectName()));

    if (!Action((TClassInfo *) Object->GetClassInfo(), Object, ActionMigrated))
        return FALSE;

    if (CheckBoundaries)
    {
        if (PreviousCPNIDOfLatestContainer != INVALID)
        {
            TExtendedCPNInfo     *Info;
            TContainerObjectEnum  CObject;
        
            if ((Info = Network->GetExtendedCPNInfo(PreviousCPNIDOfLatestContainer, TRUE)) == 0)
            {
                Error("ReceiveMigrationInfo: cannot retrieve ExtendedCPNInfo");
                return FALSE;
            }
            CObject.WalksForwardThrough(Info->BoundaryContainers);
            while (CObject())
            {
                TContainerReference  Refs[MAX_CONTAINER_DEPS];
                int                  NrOfRefs;
                int                  CRef;
                
                if (!CObject->GetRequiredContainers(NrOfRefs, Refs))
                {
                    return FALSE;
                }
                for (CRef = 0; CRef < NrOfRefs; CRef++)
                    if (Refs[CRef].ClassID == Object->GetClassID() AND
                        Refs[CRef].ParallelID == Object->GetParallelID() AND
                        Info->CPNInfo->ID != Object->GetCPNID())
                    {
                        IfDebug(DM_MIGRATION, Print("move container (%ld:%ld) from %lX to %lX",
                                                    Refs[CRef].ClassID, Refs[CRef].ParallelID, 
                                                    PreviousCPNIDOfLatestContainer, 
                                                    Object->GetCPNID()));
                        if (!AddBoundaryContainer(CObject, Object->GetCPNID()))
                        {
                            Error("HandleMigrationInfo: cannot add boundary");
                            return FALSE;
                        }
                        if (!RemoveBoundaryContainer(CObject, PreviousCPNIDOfLatestContainer))
                        {
                            Error("HandleMigrationInfo: cannot remove boundary");
                            return FALSE;
                        }
                    }
                CObject++;
            }
        }
    }
    return TRUE;
}

// =================================================================================== 

TParallelObject *
TObjectManager::ReceiveObject(TBaseStream *aStream, bool UseExisting,
                              TTransferInfo *TransferInfo)

{
    TParallelObject *Object = 0;
    TClassInfo      *ClassInfo;
    TParallelID      ParallelID;
    TClassID         ClassID;

    if (!(aStream->Decode(ClassID) AND
          aStream->Decode(ParallelID)))
    {
        Error("ReceiveObject: cannot decode ClassID and ParallelID");
        return 0;
    }
    if (ClassID < 0 OR ClassID >= ClassInfos.GetNrOfElements())
    {
        Error("ReceiveObject: invalid ClassID %d", ClassID);
        return 0;
    }
    ClassInfo = GetClassInfo(ClassID);
    if (UseExisting)
    {
        if (ClassInfo->Table == 0)
        {
            Error("ReceiveObject: cannot use existing instance for (%s)", 
                  (char *) ClassInfo->GetObjectName());
            return 0;
        }
        Object = ClassInfo->GetObjectPtr(ParallelID);
    }
    if (Object == 0)
    {
        PreviousObjectExisted = FALSE;
        if ((Object = CreateObject(ClassInfo, ParallelID)) == 0)
        {
            Error("ReceiveObject: cannot create %s", (char *) ClassInfo->GetInstanceName());
            return 0;
        }
    }
    else
        PreviousObjectExisted = TRUE;
    PreviousCPNIDOfLatestContainer = Object->GetCPNID();
    if (!Object->Decode(aStream, TransferInfo))
    {
        Error("ReceiveObject: cannot decode '%s'", (char *) Object->GetObjectName());
        return 0;
    }
    return Object;
}

// =================================================================================== 

bool
TObjectManager::ReceiveObjectList(TParallelObjectList *ObjectList, 
                                  TBaseStream *aStream, bool UseExisting,
                                  TTransferInfo *TransferInfo)

{
    TParallelListObject *Object;
    long                 Count;
    long                 Header;

    Count = 0;
    if (!aStream->Decode(Header))
        return FALSE;
    while (Header)
    {
        Count++;
        if (Header != Count)
        {
            Error("ReceiveObjectList: inconsistent list item header");
            return FALSE;
        }
        if ((Object = (TParallelListObject *) 
             ReceiveObject(aStream, UseExisting, TransferInfo)) == 0)
        {
            Error("ReceiveObjectList: cannot receive list item %ld", Count);
            return FALSE;
        }
        ObjectList->Append(Object);
        if (!aStream->Decode(Header))
            return FALSE;
    }    
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::RegisterClass(int theFlags, TClassInfo *&ClassInfo, 
                              const char *ContainerClassName,
                              const char *ClassName, TObjectHandler *ObjectHandler)

{
    ClassInfo = new TClassInfo(theFlags, ContainerClassName, ClassName, ObjectHandler);
    if (!ClassInfos.Insert(ClassInfo, TRUE))
    {
        Error("class name '%s' already registered", ClassName);
        return FALSE;
    }
    debug(Print(DM_CLASS, "register class '%s'", (char *) ClassInfo->GetInstanceName()));
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::RegisterContainerClass(int Flags, TClassInfo *&ClassInfo,
                                       const char *ClassName, TObjectHandler *ObjectHandler)

{
    Flags |= CI_FLAG_CONTAINER_CLASS | CI_FLAG_TABLE;
    ClassInfo = new TClassInfo(Flags, 0, ClassName, ObjectHandler);
    if (!ClassInfos.Insert(ClassInfo, TRUE))
    {
        Error("class name '%s' already registered", ClassName);
        return FALSE;
    }
    debug(Print(DM_CLASS, "register container class '%s'", 
                (char *) ClassInfo->GetInstanceName()));
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::RemoveBoundaryContainer(TContainerObject *Object, TCPNID CPNID)

{
    TExtendedCPNInfo *Info;

    if ((Info = Network->GetExtendedCPNInfo(CPNID, TRUE)) == 0)
    {
        Error("RemoveBoundaryContainer: cannot add ExtendedCPNInfo "
              "for boundary %s", (char *) Object->GetObjectName());
        return FALSE;
    }
    if (Info->BoundaryContainers.IsInList(Object))
    {
        IfDebug(DM_TOPOLOGY, Print("rem bnd cntnr %s from %lX",
                                   (char *) Object->GetObjectName(), Info->CPNInfo->ID));
        Info->BoundaryContainers.Delete(Object);
        if (!Action(Object->GetClassInfo(), Object, ActionRemoveBoundary))
            return FALSE;
        if (Info->BoundaryContainers.GetNrOfElements() == 0 AND
            Info->CPNInfo != Parallel->GetLocalCPNInfo() AND
            Neighbors.IsInList(Info->CPNInfo))
        {
            IfDebug(DM_TOPOLOGY, Print("rem nghbr %lX", Info->CPNInfo->ID));
            Neighbors.Delete(Info->CPNInfo);
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::RemoveContainer(TContainerObject *Object)

{
    if (LocalContainerObjects.IsInList(Object))
    {
        IfDebug(DM_CONTAINER, Print("RemoveContainer: %s", (char *) Object->GetObjectName()));  
        if (!Action(Object->GetClassInfo(), Object, ActionRemove))
            return FALSE;
        LocalContainerObjects.Delete(Object);   
        NewContainerObjects.Delete(Object);
        Object->DisposeNonLocalInfos();
        DestroyInstance(Object);
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::RemoveNonLocalInstances(void)

{
    TContainerObjectEnum CObject;
    int                  Count;
    int                  PotentialCount;

    Count = 0;
    PotentialCount = (int) TransferContainers.GetNrOfElements();

    // Walk through all dependant objects and check which required objects are local. 
    // If a required object is found that is not local it will be added to the 
    // RemoteContainers list.

    CObject.WalksForwardThrough(DependantContainers);
    while (CObject())
    {
        TContainerReference  Refs[MAX_CONTAINER_DEPS];
        int  NrOfRefs;    
        int  CRef;
        bool IsPartiallyLocal = FALSE;
        bool IsLocal = TRUE;
                
        if (!CObject->GetRequiredContainers(NrOfRefs, Refs + 0))
        {
            Error("CheckNewBoundaries: cannot retrieve refs for %s",
                  (char *) CObject->GetObjectName());
            return FALSE;
        }
        for (CRef = 0; CRef < NrOfRefs; CRef++)
        {
            TContainerObject *RequiredContainer;
            
            if ((RequiredContainer = GetContainer(Refs[CRef])) == 0)
            {
                Error("CheckNewBoundaries: cannot retrieve required container %d:%ld",
                      Refs[CRef].ClassID, Refs[CRef].ParallelID);
                return FALSE;
            }
            if (RequiredContainer->IsLocal(Parallel->GetLocalCPNID()))
                IsPartiallyLocal = TRUE;
            else
                IsLocal = FALSE;
        }
        if (!IsLocal)
        {
            if (IsPartiallyLocal)
            {
                for (CRef = 0; CRef < NrOfRefs; CRef++)
                {
                    TContainerObject *RequiredContainer;
                    
                    if ((RequiredContainer = GetContainer(Refs[CRef])) == 0)
                    {
                        Error("RemoveNonLocalInstances: cannot retrieve required container");
                        return FALSE;
                    }
                    if (!RequiredContainer->IsLocal(Parallel->GetLocalCPNID()))
                    {
                        AddRemoteContainer(RequiredContainer);
                        if (!AddBoundaryContainer(CObject, RequiredContainer->GetCPNID()))
                        {
                            Error("RemoveNonLocalInstances: cannot add boundary");
                            return FALSE;
                        }
                    }
                }
                if (!CObject->DisposeNonLocalPortion(Parallel->GetLocalCPNID()))
                {
                    Error("RemoveNonLocalInstances: cannot dispose non local portion for %s",
                          (char *) CObject->GetObjectName());
                    return FALSE;
                }
                IfDebug(DM_MIGRATION, Print("split dependant %s", 
                                            (char *) CObject->GetObjectName()));
                if (!Action((TClassInfo *) CObject->GetClassInfo(), CObject, ActionSplit))
                    return FALSE;               
            }
            else
            {
                IfDebug(DM_MIGRATION, Print("migrate dependant %s", 
                                            (char *) CObject->GetObjectName()));
                ContainersToBeRemoved.Append(CObject);
            }
        }
        CObject++;
    }
    CObject.WalksForwardThrough(ContainersToBeRemoved);
    while (CObject())
    {
        if (!RemoveContainer(CObject))
        {
            Error("RemoveNonLocalContainers: cannot remove container");
            return FALSE;
        }
        CObject++;
    }
    CObject.WalksForwardThrough(RequiredContainers);
    while (CObject())
    {
        if (!CObject->IsLocal(Parallel->GetLocalCPNID()))
        {
            IfDebug(DM_MIGRATION, Print("migrate %s", (char *) CObject->GetObjectName()));
            if (!RemoveContainer(CObject))
            {
                Error("RemoveNonLocalInstances: cannot remove %s",
                      (char *) CObject->GetObjectName());
                return FALSE;
            }
            Count++;
        }
        CObject++;
    }
    Print(DM_CONTAINER, "RemoveNonLocalInstances: %ld of potentially %ld objects", 
          Count, PotentialCount);
    TransferContainers.DeleteAll();
    RequiredContainers.DeleteAll();
    DependantContainers.DeleteAll();
    ContainersToBeRemoved.DeleteAll();
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::RemoveOldBoundaries(TCPNID FromCPNID)

{
    TContainerObject     *RequiredContainer;
    TContainerObjectEnum  CObject;
    TContainerReference   Refs[MAX_CONTAINER_DEPS];
    int                   NrOfRefs;
    int                   CRef;
    TExtendedCPNInfo     *Info;
    
    TContainerObjectList  ToBeDeleted;

    if ((Info = Network->GetExtendedCPNInfo(FromCPNID, TRUE)) == 0)
    {
        Error("RemoveOldBoundaries: cannot get ExtendedCPNInfo for %lX", FromCPNID);
        return FALSE;
    }   
    CObject.WalksForwardThrough(Info->BoundaryContainers);
    while (CObject())
    {
        bool HasLocalReq = FALSE;
        bool HasNonLocalReq = FALSE;

        if (!CObject->GetRequiredContainers(NrOfRefs, Refs + 0))
        {
            Error("RemoveOldBoundaries: cannot get required containers for %s",
                  (char *) CObject->GetObjectName());
            return FALSE;
        }
        if (NrOfRefs == 0)
            Warning("******* NUMBER OF REFS = 0 for %s", (char *) CObject->GetObjectName());
        for (CRef = 0; CRef < NrOfRefs; CRef++)
        {
            if ((RequiredContainer = GetContainer(Refs[CRef])) == 0)
            {
                Error("InitDataAfterReception: cannot retrieve required container");
                return FALSE;
            }
            if (RequiredContainer->IsLocal(Parallel->GetLocalCPNID()))
                HasLocalReq = TRUE;
            else
                HasNonLocalReq = TRUE;
        }
        if (HasLocalReq != HasNonLocalReq)
            ToBeDeleted.Append(CObject);
        CObject++;
    }
    CObject.WalksForwardThrough(ToBeDeleted);
    while (CObject())
    {   
        if (!RemoveBoundaryContainer(CObject, FromCPNID))
        {
            Error("InitDataAfterReception: cannot remove boundary container");
            return FALSE;
        }
        CObject++;
    }    
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::RemoveRemoteContainer(TContainerObject *Object)

{
    if (RemoteContainers.IsInList(Object))
    {
        IfDebug(DM_MIGRATION, Print("remove remote container %s",
                                    (char *) Object->GetObjectName()));
        RemoteContainers.Delete(Object);
        Object->RemoveInstance();       
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::Run(void)

{
    if (!ActivateClasses())
        return FALSE;
    if (Parallel)
        Neighbors.Append(Parallel->GetLocalCPNInfo());
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::SendBoundaries(TTimeStep &TimeStep, TTransferInfo *TransferInfo)

{
    TEnumerator< TExtendedCPNInfo > CCPN(TExtendedCPNInfo::ClassInfo->GetTable());

    if (BoundaryTimeStep != (long) 0 AND
        BoundaryTimeStep != TimeStep)
    {
        Error("SendBoundaries: discontinuous time-steps");
        return FALSE;
    }
    BoundaryTimeStep = TimeStep;
    Print(DM_BOUNDARY, "SendBoundaries: time-step %s (missing sets:%ld)", 
          (char *) TimeStep.GetInstanceName(), MissingBoundarySets[0]);
    CCPN.Enumerate();
    while(CCPN())
    {
        if (CCPN->BoundaryContainers.GetNrOfElements())
        {
            TBaseMessage         *Message;
            TContainerObjectEnum  CContainer;
            TBoundaryList         Boundaries;

            CContainer.WalksForwardThrough(CCPN->BoundaryContainers);
            while (CContainer())
            {
                if (!CContainer->GetBoundaries(BoundaryTimeStep, &Boundaries))
                {
                    Error("cannot get boundaries from %s", 
                          (char *) CContainer->GetInstanceName());
                    return FALSE;
                }
                CContainer++;
            }
            if (Boundaries.GetNrOfElements() > 0)
            {
                if (!Communication->GetNewMessage(Message, CCPN->CPNInfo->ID, 
                                                  MTAG_BOUNDARIES, MTYPE_OBJ_MGR))
                {
                    Error("SendBoundaries: cannot get message");
                    return FALSE;
                }
                debug(Print(DM_BOUNDARY, "sending %ld boundaries to %lX", 
                            Boundaries.GetNrOfElements(), CCPN->CPNInfo->ID));
                if (!(Message->Event(MEVENT_OBJ_MGR_BOUNDARIES) AND
                      TimeStep.Encode(Message) AND
                      TransferInfo->Encode(Message) AND
                      SendObjectList((TParallelObjectList *) &Boundaries, 
                                     Message, TransferInfo) AND
                      Communication->Commit(Message)))
                {
                    Error("cannot send boundaries");
                    return FALSE;
                }
                Boundaries.DeleteAll(TRUE);
                MissingBoundarySets[0]++;
            }
        }
        CCPN++;
    }    
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::SendMigrationInfos(void)

{
    TContainerObjectEnum  CObject;
    TContainerReference   Refs[MAX_CONTAINER_DEPS];
    TTransferInfo         TransferInfo;
    TBaseMessage         *GlobalMessage = 0;

    Print(DM_CONTROL, "send migration infos");
    TransferInfo.TransferMode = TM_MigrationInfo;
    if (GlobalMigrationMonitor != INVALID)
    {       
        if (!Communication->GetNewMessage(GlobalMessage, GlobalMigrationMonitor,
                                          MTAG_OBJ_MGR, MTYPE_OBJ_MGR_MIGRATION_INFOS))
        {
            Error("SendRemoteInfos: cannot get message to %lX",
                  GlobalMigrationMonitor);
            return FALSE;
        }
    }
    TransferInfo.TransferMode = TM_MigrationInfo;
    CObject.WalksForwardThrough(RequiredContainers);
    while (CObject())
    {
        int NrOfRefs;
        int CRef;

        if (GlobalMessage)
            if (!(GlobalMessage->Select() AND
                  GlobalMessage->Event(MEVENT_OBJ_MGR_MONITOR_MIGRATION_INFO) AND
                  SendObject((TContainerObject *) CObject, GlobalMessage, &TransferInfo)))
            {
                Error("SendMigrationInfos: cannot encode migration info for %s",
                      (char *) CObject->GetObjectName());
                return FALSE;
            }
        if (!CObject->GetDependantContainers(NrOfRefs, Refs))
        {
            Error("SendMigrationInfos: cannot retrieve dependant containers for %s",
                  (char *) CObject->GetObjectName());
            return FALSE;
        }
        for (CRef = 0; CRef < NrOfRefs; CRef++)
        {
            TContainerObject     *Dependant;
            TContainerReference   Refs2[MAX_CONTAINER_DEPS];
            int                   NrOfRefs2;
            int                   CRef2;
            
            if ((Dependant = GetContainer(Refs[CRef])) == 0)
            {
                Error("SendMigrationInfos: cannot retrieve dependant container %d:%ld",
                      Refs[CRef].ClassID, Refs[CRef].ParallelID);
                return FALSE;
            }
            if (!Dependant->GetRequiredContainers(NrOfRefs2, Refs2))
            {
                Error("SendMigrationInfos: cannot retrieve required containers for %s",
                      (char *) Dependant->GetObjectName());
                return FALSE;
            }
            for (CRef2 = 0; CRef2 < NrOfRefs2; CRef2++)
            {
                TContainerObject *Required;
                
                if ((Required = GetContainer(Refs2[CRef2])) == 0)
                {
                    Error("SendMigrationInfos: cannot retrieve required container %d:%ld",
                          Refs2[CRef2].ClassID, Refs2[CRef2].ParallelID);
                    return FALSE;
                }
                if (Required->GetCPNID() != CObject->GetCPNID() AND
                    Required->GetCPNID() != Parallel->GetLocalCPNID())
                {
                    TBaseMessage *Message;
                    TCPNInfo     *Info;
                    
                    if (!(Communication->GetNewMessage(Message, Required->GetCPNID(),
                                                       MTAG_MIGRATION, 
                                                       MTYPE_OBJ_MGR_MIGRATION_INFOS) AND
                          Message->Event(MEVENT_OBJ_MGR_MIGRATION_INFO) AND
                          SendObject((TContainerObject *) CObject, Message, &TransferInfo) AND
                          Communication->Commit(Message)))
                    {
                        Error("SendMigrationInfo: cannot send message to %lX",
                              Required->GetCPNID());
                        return FALSE;
                    }
                    IfDebug(DM_MIGRATION, Print("send migration info for %s to %lX",
                                                (char *) CObject->GetObjectName(), 
                                                Required->GetCPNID()));
                    if (!(Communication->GetNewMessage(Message, CObject->GetCPNID(),
                                                       MTAG_MIGRATION, 
                                                       MTYPE_OBJ_MGR_MIGRATION_INFOS) AND
                          Message->Event(MEVENT_OBJ_MGR_MIGRATION_INFO) AND
                          SendObject(Required, Message, &TransferInfo) AND
                          Communication->Commit(Message)))
                    {
                        Error("SendMigrationInfo: cannot send message to %lX",
                              Required->GetCPNID());
                        return FALSE;
                    }
                    IfDebug(DM_MIGRATION, Print("send migration info for %s to %lX",
                                                (char *) Required->GetObjectName(), 
                                                CObject->GetCPNID()));
                    if ((Info = Parallel->GetCPNInfoByID(Required->GetCPNID())) == 0)
                    {
                        Error("SendMigrationInfos: cannot retrieve CPNInfo");
                        return FALSE;
                    }
                    TransferDestinations.Append(Info);
                }
            }
        }
        CObject++;
    }
    if (GlobalMessage)
        if (!Communication->Commit(GlobalMessage))
        {
            Error("SendMigrationInfo: cannot send message to global monitor %lX",
                  GlobalMigrationMonitor);
            return FALSE;
        }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::SendObject(const TParallelObject *Object, TBaseStream *aStream, 
                           TTransferInfo *TransferInfo) const

{
    TClassID    ClassID;
    TParallelID ParallelID;

    ClassID = Object->GetClassID();
    ParallelID = Object->GetParallelID();
    if (!(aStream->Encode(ClassID) AND
          aStream->Encode(ParallelID) AND
          Object->Encode(aStream, TransferInfo)))
    {
        Error("EncodeObject: cannot encode '%s'", (char *) Object->GetInstanceName());
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TObjectManager::SendObjectList(const TParallelObjectList *ObjectList, TBaseStream *aStream, 
                               TTransferInfo *TransferInfo) const

{
    TParallelObjectEnum CObject;
    long                Count;

    Count = 0;
    CObject.WalksForwardThrough(ObjectList);
    while (CObject())
    {
        if ((TransferInfo == 0) OR
            (TransferInfo != 0) AND CObject->IsInRange(&TransferInfo->TransferRange))
        {
            Count++;
            if (!(aStream->Encode(Count) AND
                  SendObject((TParallelObject *) CObject, aStream, TransferInfo)))
            {
                Error("EncodeObjectList: cannot encode '%s'", 
                      (char *) CObject->GetInstanceName());
                return FALSE;
            }
        }
        CObject++;
    }
    Count = 0;
    return aStream->Encode(Count);
}

// =================================================================================== 

bool
TObjectManager::Wait(TCPNID Sender)

{
    if (Sender != Parallel->GetLocalCPNID())
    {
        Print(DM_MIGRATION, "waiting for topology/migration infos");
        Communication->ActivateTags(MTAG_MIGRATION);
        DataReceived = FALSE;
        while (!DataReceived)
            if (!Communication->MessageLoop(TRUE, Sender))
            {
                Error("Wait: while waiting for topology/migration infos");
                return FALSE;
            }
        Communication->DeactivateTags(MTAG_MIGRATION);
        Print(DM_MIGRATION, "finished waiting");
    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
