/*      file     : xparallel/TTopologyWindow.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TTopologyWindow.C,v 1.9 2005/03/25 17:01:28 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <math.h>

#include <Graphx.h>
#include <XGraphx.h>
#include <TWindow.h>
#include <TConfiguration.h>
#include <TNetwork.h>
#include <TParallel.h>
#include <TTopology.h>
#include <TExtendedCPNInfo.h>
#include <TSamplingDescription.h>
#include <TAggregationInfo.h>
#include <TAggregationResult.h>
#include <TStatManager.h>
#include <TExtendedLoadStat.h>
#include <TCommunicationLink.h>
#include <TNeighborCommInfo.h>

#include "XParallelConfig.h"
#include "TTopologyWindow.h"

// ***DEFINES************************************************************************* 

#define TOPOWIN_DEFAULT_NAME                   "Topology"
#define TOPOWIN_DEFAULT_BORDER                 0.1
#define TOPOWIN_DEFAULT_WIDTH                  300L
#define TOPOWIN_DEFAULT_HEIGHT                 300L
#define TOPOWIN_DEFAULT_RELATIVERADIUS         0.2
#define TOPOWIN_DEFAULT_RELATIVEARROWWIDTH     0.2
#define TOPOWIN_DEFAULT_RELATIVEBNDARROWWIDTH  0.01
#define TOPOWIN_DEFAULT_BACKGROUNDCOLOR        220L
#define TOPOWIN_DEFAULT_IDLETIMECOLOR          GR_RED
#define TOPOWIN_DEFAULT_MISCTIMECOLOR          GR_BLUE
#define TOPOWIN_DEFAULT_STATTIMECOLOR          GR_YELLOW
#define TOPOWIN_DEFAULT_APPTIMECOLOR           GR_GREEN
#define TOPOWIN_DEFAULT_SAMPLINGINTERVAL       10L

#define TOPOWIN_RESERVED_COLORS 8

#define TOPOWIN_FLAG_SHOW_LINKS             1
#define TOPOWIN_FLAG_SHOW_LOAD_TRANSFER     2

#define TOPOWIN_FLAG_DEFAULT (TOPOWIN_FLAG_SHOW_LINKS | TOPOWIN_FLAG_SHOW_LOAD_TRANSFER)

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TWindow
// =================================================================================== 

TTopologyWindow::TTopologyWindow(TTopology *theTopology)

{
    Topology = theTopology;
    ApplicationLoad = 0.0;
    Window = 0;
    FrameAvailable = FALSE;
    StatChanged = TRUE;
    DrawFlags = TOPOWIN_FLAG_DEFAULT;
    Name = Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_NAME)? 
        Configuration->GetString() : TOPOWIN_DEFAULT_NAME;
    Width = Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_WIDTH)? 
        Configuration->GetLong() : TOPOWIN_DEFAULT_WIDTH;
    Height = Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_HEIGHT)? 
        Configuration->GetLong() : TOPOWIN_DEFAULT_HEIGHT;
    Border = Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_BORDER)? 
        Configuration->GetDouble() : TOPOWIN_DEFAULT_BORDER;
    RelativeNodeRadius = Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_RELATIVERADIUS)? 
        Configuration->GetDouble() : TOPOWIN_DEFAULT_RELATIVERADIUS;
    RelativeArrowWidth = Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_RELATIVEARROWWIDTH)? 
        Configuration->GetDouble() : TOPOWIN_DEFAULT_RELATIVEARROWWIDTH;
    RelativeBndArrowWidth = Configuration->Exists(TOPOWIN_GROUP, 
                                                  TOPOWIN_NAME_RELATIVEBNDARROWWIDTH)? 
        Configuration->GetDouble() : TOPOWIN_DEFAULT_RELATIVEBNDARROWWIDTH;
    BackgroundColor = (int) (Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_BACKGROUNDCOLOR)? 
                             Configuration->GetLong() : TOPOWIN_DEFAULT_BACKGROUNDCOLOR);
    IdleTimeColor = (int) (Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_IDLETIMECOLOR)? 
                           Configuration->GetLong() : TOPOWIN_DEFAULT_IDLETIMECOLOR);
    MiscTimeColor = (int) (Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_MISCTIMECOLOR)? 
                           Configuration->GetLong() : TOPOWIN_DEFAULT_MISCTIMECOLOR);
    AppTimeColor = (int) (Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_APPTIMECOLOR)? 
                          Configuration->GetLong() : TOPOWIN_DEFAULT_APPTIMECOLOR);
    StatTimeColor = (int) (Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_STATTIMECOLOR)? 
                           Configuration->GetLong() : TOPOWIN_DEFAULT_STATTIMECOLOR);
    SamplingInterval = (int) (Configuration->Exists(TOPOWIN_GROUP, TOPOWIN_NAME_SAMPLINGINTERVAL)? 
                              Configuration->GetLong() : TOPOWIN_DEFAULT_SAMPLINGINTERVAL);
    LoadSamplingDescription = 
        new TSamplingDescription(TExtendedLoadStat::ClassInfo->GetID(), 0,
                                 Parallel->GetLocalCPNID(), 
                                 SAMPLING_ID_EXTENDED_LOAD_STAT);
    LoadSamplingDescription->SetCollectInterval(SamplingInterval);
    LoadAggregationInfo = new TAggregationInfo(this, LoadSamplingDescription);
    StatManager->RegisterAggregation(LoadAggregationInfo);
    StatManager->StartSampling(LoadSamplingDescription, TRUE);
}

// =================================================================================== 

TTopologyWindow::~TTopologyWindow(void)

{
    DeleteObject(Window);
    /* may cause MPI to block ?????
    StatManager->StopSampling(LoadSamplingDescription);
    StatManager->CancelAggregation(LoadAggregationInfo);
    */
    DeleteObject(LoadSamplingDescription);
    DeleteObject(LoadAggregationInfo);
}

// =================================================================================== 

bool
TTopologyWindow::CheckOpenWindow(void)

{
    if (Window == 0 AND FrameAvailable AND ApplicationLoad > 0.0)
    {
        double Ratio;

        Ratio = Frame.GetDeltaX() / Frame.GetDeltaY();
        if (Ratio > Width / Height)
            Height = (int) (1.0 * Width / Ratio);
        else
            Width = (int) (Ratio * Height);
        Window = new TWindow(Name, Width, Height, TOPOWIN_RESERVED_COLORS);
        if (Window->Failed())
        {
            Error("CheckOpenWindow: cannot create topology window");
            return FALSE;
        }
        Window->SetXaxis(Frame.GetMinX(), Frame.GetMaxX());
        Window->SetYaxis(Frame.GetMinY(), Frame.GetMaxY());
        Window->SetVerticalTextAlignment(GrTopAlign);
        Window->X11SetSpeed(GrFast);
        Window->X11WaitUntilExposed();
    }
    return TRUE;
}

// =================================================================================== 

bool
TTopologyWindow::Draw(void)

{    
    TEnumerator < TExtendedCPNInfo > CInfo(TExtendedCPNInfo::ClassInfo->GetTable());
    TInfoID StatInfoID;

    if (ApplicationLoad <= 0.0)
    {
        Warning("Draw: application load < 0.0");
        return TRUE;
    }
    Window->ClearScreen(BackgroundColor);
    StatInfoID = TExtendedLoadStat::ClassInfo->GetInfoID();
    CInfo.Enumerate();
    while (CInfo())
    {
        TExtendedLoadStat *Stat;
        
        if ((Stat = (TExtendedLoadStat *) CInfo->GetInfo(StatInfoID)))
        {
            double Radius;
            double AppTimeAngle;
            double StatTimeAngle;
            double MiscTimeAngle;
            double SumTime;
            double SumAngle;

            SumTime = Stat->GetAverageApplicationTime() + Stat->GetAverageIdleTime() +
                Stat->GetAverageMiscTime();
            AppTimeAngle = 2.0 * 64 * PI * Stat->GetAverageApplicationTime() / SumTime;
            MiscTimeAngle = 2.0 * 64 * PI * Stat->GetAverageMiscTime() / SumTime;
            StatTimeAngle = 2.0 * 64 * PI * Stat->GetAverageStatTime() / SumTime;
            if (Stat->GetAverageApplicationLoad() > 0)
            {
                Radius = NodeRadius * sqrt(Stat->GetAverageApplicationLoad());

                SumAngle = 0.0;
                *Window << GrFillArc(Stat->GetCenterOfGravity().x(),
                                     Stat->GetCenterOfGravity().y(), Radius, 
                                     SumAngle, SumAngle + AppTimeAngle, 
                                     AppTimeColor);
                SumAngle += AppTimeAngle;
                *Window << GrFillArc(Stat->GetCenterOfGravity().x(),
                                     Stat->GetCenterOfGravity().y(), Radius, 
                                     SumAngle, SumAngle + StatTimeAngle, 
                                     StatTimeColor);
                SumAngle += StatTimeAngle;
                *Window << GrFillArc(Stat->GetCenterOfGravity().x(),
                                     Stat->GetCenterOfGravity().y(), Radius, 
                                     SumAngle, SumAngle + MiscTimeAngle,
                                     MiscTimeColor);
                SumAngle += MiscTimeAngle;
                *Window << GrFillArc(Stat->GetCenterOfGravity().x(),
                                     Stat->GetCenterOfGravity().y(), Radius,
                                     SumAngle, 2 * 64 * PI, 
                                     IdleTimeColor);
            }
        }           
        CInfo++;
    }
    DrawLinks();
    Window->MarkForRedisplay();
    return TRUE;
}

// =================================================================================== 

bool
TTopologyWindow::DrawLinks(void)

{
    TEnumerator< TCommunicationLink > CLink(Network->GetCommunicationLinkTable());
    TInfoID InfoID;
    TInfoID StatInfoID;
    double  MaxTransferLoad;
    long    MaxBoundaries;

    InfoID = TNeighborCommInfo::ClassInfo->GetInfoID();
    StatInfoID = TExtendedLoadStat::ClassInfo->GetInfoID();
    CLink.Enumerate();
    MaxTransferLoad = 0;
    MaxBoundaries = 0;
    while (CLink())
    {
        TNeighborCommInfo *Info;

        if ((Info = (TNeighborCommInfo *) CLink->GetInfo(InfoID)))
        {
            if (Info->TransferLoad > MaxTransferLoad)
                MaxTransferLoad = Info->TransferLoad;
            if (Info->NrOfBoundaries > MaxBoundaries)
                MaxBoundaries = Info->NrOfBoundaries;
        }
        CLink++;
    }
    CLink.Enumerate();
    while (CLink())
    {
        TNeighborCommInfo *Info;
        TExtendedCPNInfo  *FromInfo, *ToInfo = 0;
        TExtendedLoadStat *FromStat = 0, *ToStat = 0;
        
        if ((Info = (TNeighborCommInfo *) CLink->GetInfo(InfoID)))
        {
            if ((FromInfo = Network->GetExtendedCPNInfo(CLink->GetFromCPNID(), TRUE)) AND
                (ToInfo = Network->GetExtendedCPNInfo(CLink->GetToCPNID(), TRUE)) AND
                (FromStat = (TExtendedLoadStat *) FromInfo->GetInfo(StatInfoID)) AND
                (ToStat = (TExtendedLoadStat *) ToInfo->GetInfo(StatInfoID)))
            {
                if (DrawFlags & TOPOWIN_FLAG_SHOW_LOAD_TRANSFER AND Info->TransferLoad > 0.0)
                {
                    double FromRadius = NodeRadius * 
                        sqrt(FromStat->GetAverageApplicationLoad());
                    double ToRadius = NodeRadius * 
                        sqrt(ToStat->GetAverageApplicationLoad());
                    
                    Window->DrawArrow(FromStat->GetCenterOfGravity(), FromRadius,
                                      ToStat->GetCenterOfGravity(), ToRadius,
                                      ArrowWidth * sqrt(Info->TransferLoad),
                                      TRUE, GR_RED);
                }
                
                if (DrawFlags & TOPOWIN_FLAG_SHOW_LINKS AND CLink->IsActive())
                    Window->DrawArrow(FromStat->GetCenterOfGravity(), 0.0,
                                      ToStat->GetCenterOfGravity(), 0.0,
                                      BoundaryArrowWidth * 
                                      Info->NrOfBoundaries / MaxBoundaries, FALSE, 0);
            }
        }       
        CLink++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TTopologyWindow::ExecuteSubTimeStep(TTimeStep &)

{
    if (!CheckOpenWindow())
    {
        Error("ExecuteSubTimeStep: while opening window");
        return FALSE;
    }
    if (Window)
    {
        if (StatChanged)
        {
            if (!Draw())
            {
                Error("ExecuteSubTimeStep: while drawing");
                return FALSE;
            }
            StatChanged = FALSE;
        }
        if (!Window->Execute())
        {
            Error("ExecuteSubTimeStep: while executing topology window");
            return FALSE;
        }
    }
    return TRUE;
}

// =================================================================================== 

TString
TTopologyWindow::GetClassName(void) const

{
    return "TTopologyWindow";
}

// =================================================================================== 

bool
TTopologyWindow::HandleExtendedLoadResult(TAggregationInfo *Info, TTimeStep &TimeStep)

{
    TAggregationResult     *Result = 0;
    TExtendedLoadStatList  *StatList = 0;
    TExtendedLoadStatEnum   CStat;
    TInfoID                 StatInfoID;

    if ((Result = (TAggregationResult *) Info->GetResult(TimeStep)) == 0)
    {
        Error("HandleLoadResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((StatList = (TExtendedLoadStatList *) Result->StatObjectList) == 0)
    {
        Error("HandleLoadResult: cannot retrieve stat object list");
        return FALSE;
    }   
    StatInfoID = TExtendedLoadStat::ClassInfo->GetInfoID();
    Frame.Empty();
    CStat.WalksForwardThrough(StatList);
    ApplicationLoad = 0.0;
    while (CStat())
    {   
        TExtendedCPNInfo *CPNInfo;

        if ((CPNInfo = Network->GetExtendedCPNInfo(CStat->GetParallelID(), TRUE)) == 0)
        {
            Error("HandleLoadResult: cannot retrieve ExtendedCPNInfo for %lX",
                  CStat->GetParallelID());
            return FALSE;
        }
        CPNInfo->SetInfo(StatInfoID, (TExtendedLoadStat *) CStat);
        Frame.Merge(CStat->GetFrame());
        ApplicationLoad += CStat->GetAverageApplicationLoad();
        CStat++;
    }
    StatList->DeleteAll();
    Info->DisposeResult(TimeStep);
    Frame *= (1.0 + Border);
    NodeRadius = RelativeNodeRadius * Frame.GetDeltaX() / sqrt(ApplicationLoad);
    ArrowWidth = RelativeArrowWidth * Frame.GetDeltaX() / sqrt(ApplicationLoad);
    BoundaryArrowWidth = RelativeBndArrowWidth * Frame.GetDeltaX();
    StatChanged = TRUE;
    FrameAvailable = TRUE;
    return TRUE;
}

// =================================================================================== 

bool
TTopologyWindow::HandleAggregationResult(TAggregationInfo *AggregationInfo, 
                                         TTimeStep &TimeStep)

{
    if (AggregationInfo == LoadAggregationInfo)
        return HandleExtendedLoadResult(AggregationInfo, TimeStep);
    return TStatHandler::HandleAggregationResult(AggregationInfo, TimeStep);
}

// =================================================================================== 
#undef ParentClass
